'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

# pylint: disable=W0201,E1101
 
from taskcoachlib.domain.date import Date, parseDate, DateTime, parseDateTime, Recurrence

from taskcoachlib.domain.category import Category
from taskcoachlib.domain.task import Task
from taskcoachlib.domain.effort import Effort

from taskcoachlib.i18n import _

from twisted.internet.protocol import Protocol, ServerFactory
from twisted.internet.error import CannotListenError

import wx, struct, \
    random, time, hashlib, cStringIO, socket, os

# Default port is 8001.
#
# Integers are sent as 32 bits signed, network byte order.
# Strings are sent as their length (as integer), then data (UTF-8
# encoded). The length is computed after encoding.
# Dates are sent as strings, formatted YYYY-MM-DD.
#
# The exact workflow for both desktop and device is documented as Dia
# diagrams, in the "Design" subdirectory of the iPhone sources.

###############################################################################
#{ Support classes: object serialisation & packing


class BaseItem(object):
    """This is the base class of the network packet system. Each
    subclass maps to a particular type of data.

    @ivar state: convenience instance variable which starts as 0, used
        in subclasses to implement simple FSA."""

    def __init__(self):
        super(BaseItem, self).__init__()

        self.start()

    def start(self):
        """This method should reinitialize the instance."""
        self.state = 0
        self.value = None

    def expect(self):
        """This should return the number of bytes that are needed
        next. When this much bytes are finally available, they'll be
        passed to L{feed}. Return None if you're finished."""

        raise NotImplementedError

    def feed(self, data):
        """The bytes requested from L{expect} are available ('data'
        parameter)."""

        raise NotImplementedError

    def pack(self, value):
        """Unserialization. This should return a byte buffer
        representing 'value'."""

        raise NotImplementedError


class IntegerItem(BaseItem):
    """Integers. Packed as 32-bits, signed, big endian. Underlying
    type: int."""

    def expect(self):
        if self.state == 0:
            return 4
        else:
            return None

    def feed(self, data):
        self.value, = struct.unpack('!i', data)
        self.state = 1

    def pack(self, value):
        return struct.pack('!i', value)


class DataItem(BaseItem):
    """A bunch of bytes, the count being well known"""

    def __init__(self, count):
        super(DataItem, self).__init__()

        self.__count = count

    def expect(self):
        return self.__count if self.state == 0 else None

    def feed(self, data):
        if self.state == 0:
            self.value = data
            self.state = 1

    def pack(self, value):
        return value


class StringItem(BaseItem):
    """Strings. Encoded in UTF-8. Packed as their length (encoded),
    then the data. Underlying type: unicode."""

    def expect(self):
        if self.state == 0:
            return 4
        elif self.state == 1:
            return self.length
        else:
            return None

    def feed(self, data):
        if self.state == 0:
            self.length, = struct.unpack('!i', data)
            if self.length:
                self.state = 1
            else:
                self.value = u''
                self.state = 2
        elif self.state == 1:
            self.value = data.decode('UTF-8')
            self.state = 2

    def pack(self, value):
        v = value.encode('UTF-8')
        return struct.pack('!i', len(v)) + v


class FixedSizeStringItem(StringItem):
    """Same as L{StringItem}, but cannot be empty. Underlying type:
    unicode or NoneType."""

    def feed(self, data):
        super(FixedSizeStringItem, self).feed(data)

        if self.state == 2:
            if not self.value:
                self.value = None

    def pack(self, value):
        if value is None:
            return struct.pack('!i', 0)
        return super(FixedSizeStringItem, self).pack(value)


class DateItem(FixedSizeStringItem):
    """Date, in YYYY-MM-DD format. Underlying type:
    taskcoachlib.domain.date.Date."""

    def feed(self, data):
        super(DateItem, self).feed(data)

        if self.state == 2:
            self.value = Date() if self.value is None else parseDate(self.value)

    def pack(self, value):
        if isinstance(value, DateTime):
            value = Date(value.year, value.month, value.day)

        value = None if value == Date() else value.isoformat()
        return super(DateItem, self).pack(value)


class DateTimeItem(FixedSizeStringItem):
    """Date and time, YYYY-MM-DD HH:MM:SS"""

    def feed(self, data):
        super(DateTimeItem, self).feed(data)

        if self.state == 2:
            if self.value is not None:
                self.value = parseDateTime(self.value)

    def pack(self, value):
        if value is not None:
            value = value.replace(microsecond=0, tzinfo=None).isoformat(sep=' ')
        return super(DateTimeItem, self).pack(value)


class InfiniteDateTimeItem(FixedSizeStringItem):
    """Same as L{DateTimeItem}, but 'no date' is a DateTime() value
    instead of None."""

    def feed(self, data):
        super(InfiniteDateTimeItem, self).feed(data)

        if self.state == 2:
            if self.value is None:
                self.value = DateTime()
            else:
                self.value = parseDateTime(self.value)

    def pack(self, value):
        if value == DateTime():
            value = None
        if value is not None:
            value = value.replace(microsecond=0, tzinfo=None).isoformat(sep=' ')
        return super(InfiniteDateTimeItem, self).pack(value)


class CompositeItem(BaseItem):
    """A succession of several types. Underlying type: tuple. An
    exception is made if there is only 1 child, the type is then the
    same as it."""

    def __init__(self, items, *args, **kwargs):
        self._items = items

        super(CompositeItem, self).__init__(*args, **kwargs)

    def append(self, item):
        self._items.append(item)

    def start(self):
        super(CompositeItem, self).start()

        self.value = []

        for item in self._items:
            item.start()

    def expect(self):
        if self.state < len(self._items):
            expect = self._items[self.state].expect()

            if expect is None:
                self.value.append(self._items[self.state].value)
                self.state += 1
                return self.expect()

            return expect
        else:
            self.value = self.value[0] if len(self._items) == 1 else tuple(self.value)
            return None

    def feed(self, data):
        self._items[self.state].feed(data)

    def pack(self, *values):
        if len(self._items) == 1:
            return self._items[0].pack(values[0])
        else:
            return ''.join([self._items[idx].pack(v) \
                            for idx, v in enumerate(values)])

    def __str__(self):
        return 'CompositeItem([%s])' % ', '.join(map(str, self._items)) # pylint: disable=W0141


class ListItem(BaseItem):
    """A list of items. Underlying type: list."""

    def __init__(self, item, *args, **kwargs):
        self._item = item

        super(ListItem, self).__init__(*args, **kwargs)

    def start(self):
        super(ListItem, self).start()

        self.value = []

        self._item.start()

    def append(self, item):
        self._item.append(item)

    def expect(self):
        if self.state == 0:
            return 4
        elif self.state == 1:
            expect = self._item.expect()

            if expect is None:
                self.value.append(self._item.value)
                self.__count -= 1
                if self.__count == 0:
                    return None
                self._item.start()
                return self.expect()
            else:
                return expect
        elif self.state == 2:
            return None

    def feed(self, data):
        if self.state == 0:
            self.__count, = struct.unpack('!i', data)
            if self.__count:
                self._item.start()
                self.state = 1
            else:
                self.state = 2
        elif self.state == 1:
            self._item.feed(data)

    def pack(self, value):
        return struct.pack('!i', len(value)) + \
               ''.join([self._item.pack(v) for v in value])

    def __str__(self):
        return 'ListItem(%s)' % str(self._item)


class ItemParser(object):
    """Utility to avoid instantiating the Item classes by
    hand. parse('is[zi]') will hold a CompositeItem([IntegerItem(),
    StringItem(), ListItem(CompositeItem([FixedSizeStringItem(),
    IntegerITem()]))])."""

    # Special case for DataItem.

    formatMap = { 'i': IntegerItem,
                  's': StringItem,
                  'z': FixedSizeStringItem,
                  'd': DateItem,
                  't': DateTimeItem,
                  'f': InfiniteDateTimeItem }

    def __init__(self):
        super(ItemParser, self).__init__()

    @classmethod
    def registerItemType(klass, character, itemClass):
        """Register a new type of item. 'character' must be a
        single-character string, not already associated with an
        item. 'itemClass' should be a L{BaseItem} subclass. Its
        constructor must not take any parameter."""

        if len(character) != 1:
            raise ValueError('character must be a single character, not "%s".' % character)

        if character in klass.formatMap:
            raise ValueError('"%s" is already registered.' % character)

        klass.formatMap[character] = itemClass

    def parse(self, format): # pylint: disable=W0622
        if format.startswith('['):
            return ListItem(self.parse(format[1:-1]))

        current = CompositeItem([])
        stack = []
        count = None

        for character in format:
            if character == '[':
                item = ListItem(CompositeItem([]))
                stack.append(current)
                current.append(item)
                current = item
            elif character == ']':
                current = stack.pop()
            elif character == 'b':
                if count is None:
                    raise ValueError('Wrong format string: %s' % format)
                current.append(DataItem(count))
                count = None
            elif character.isdigit():
                if count is None:
                    count = int(character)
                else:
                    count *= 10
                    count += int(character)
            else:
                current.append(self.formatMap[character]())

        assert len(stack) == 0

        return current


class State(object):
    def __init__(self, disp):
        super(State, self).__init__()

        self.__disp = disp

    def init(self, format, count): # pylint: disable=W0622
        self.__format = format
        self.__count = count

        self.__data = cStringIO.StringIO()

        if format is None:
            self.__item = None
        else:
            self.__item = ItemParser().parse(format)

            if self.__count == 0:
                self.finished()
            else:
                self.__disp.set_terminator(self.__item.expect())

    def setState(self, klass, *args, **kwargs):
        self.__class__ = klass
        self.init(*args, **kwargs)

    def data(self):
        return self.__data.getvalue()

    def disp(self):
        return self.__disp

    def collect_incoming_data(self, data):
        if self.__format is not None:
            self.__data.write(data)

    def found_terminator(self):
        if self.__format is not None:
            self.__item.feed(self.__data.getvalue())
            self.__data = cStringIO.StringIO()

            length = self.__item.expect()
            if length is None:
                value = self.__item.value

                self.__count -= 1
                if self.__count:
                    self.__item.start()
                    self.__disp.set_terminator(self.__item.expect())

                self.handleNewObject(value)

                if not self.__count:
                    self.finished()
            else:
                self.__disp.set_terminator(length)

    def pack(self, format, *values):  # pylint: disable=W0622
        """Send a value."""

        self.__disp.push(ItemParser().parse(format).pack(*values))

    def handleClose(self):
        pass

    def handleNewObject(self, obj):
        raise NotImplementedError

    def finished(self):
        raise NotImplementedError

###############################################################################
# Actual protocol

_PROTOVERSION = 5


class IPhoneHandler(Protocol):
    def __init__(self):
        self.state = None
        self.__buffer = ''
        self.__expecting = None
        random.seed(time.time())

    def connectionMade(self):
        self.transport.socket.setsockopt(socket.SOL_TCP, socket.TCP_NODELAY, 1)
        self.state = BaseState(self)
        self.state.setState(InitialState, _PROTOVERSION)

    def log(self, msg, *args):
        if self.state.ui is not None:
            self.state.ui.AddLogLine(msg % args)

    def _flush(self):
        while self.__expecting is not None and len(self.__buffer) >= self.__expecting:
            data = self.__buffer[:self.__expecting]
            self.__buffer = self.__buffer[self.__expecting:]
            self.state.collect_incoming_data(data)
            self.state.found_terminator()

    def set_terminator(self, terminator):
        self.__expecting = terminator
        self._flush()

    def close_when_done(self):
        # XXX: without this delay, the other side sometimes doesn't "notice" the socket has been
        # closed... I should take a look with Wireshark...
        from twisted.internet import reactor
        reactor.callLater(0.5, self.transport.loseConnection)

    def dataReceived(self, data):
        self.__buffer += data
        self._flush()

    def connectionLost(self, reason):
        self.state.handleClose()

    def push(self, data):
        self.transport.write(data)


class IPhoneAcceptor(ServerFactory):
    protocol = IPhoneHandler

    def __init__(self, window, settings, iocontroller):
        from twisted.internet import reactor

        self.window = window
        self.settings = settings
        self.iocontroller = iocontroller

        for port in xrange(4096, 8192):
            try:
                self.__listening = reactor.listenTCP(port, self, backlog=5)
            except CannotListenError:
                pass
            else:
                break
        else:
            raise RuntimeError('Could not find a port to bind to.')

        self.port = port

    def buildProtocol(self, addr):
        password = self.settings.get('iphone', 'password')
        if password:
            protocol = ServerFactory.buildProtocol(self, addr)
            protocol.window = self.window
            protocol.settings = self.settings
            protocol.iocontroller = self.iocontroller
            return protocol

        wx.MessageBox(_('''An iPhone or iPod Touch tried to connect to Task Coach,\n'''
                        '''but no password is set. Please set a password in the\n'''
                        '''iPhone section of the configuration and try again.'''),
                        _('Error'), wx.OK)

        return None

    def close(self):
        self.__listening.stopListening()
        self.__listening = None


class BaseState(State): # pylint: disable=W0223
    def __init__(self, disp, *args, **kwargs):
        self.oldTasks = disp.window.taskFile.tasks().copy()
        self.oldCategories = disp.window.taskFile.categories().copy()

        self.ui = None

        self.syncCompleted = disp.settings.getboolean('iphone', 'synccompleted')

        super(BaseState, self).__init__(disp, *args, **kwargs)

    def isTaskEligible(self, task):
        """Returns True if a task should be considered when syncing with an iPhone/iPod Touch
        device. Right now, a task is eligible if

         * It's a leaf task (no children)
         * Or it has a reminder
         * Or it's overdue
         * Or it belongs to a category named 'iPhone'

         This will probably be more configurable in the future."""

        if task.completed() and not self.syncCompleted:
            return False

        if task.isDeleted():
            return False

        if len(task.children()) == 0:
            return True

        if task.reminder() is not None:
            return True

        if task.overdue():
            return True

        for category in task.categories():
            if category.subject() == 'iPhone':
                return True

        return False

    def handleClose(self):
        if self.ui is not None:
            self.ui.Finished()

        # Rollback
        self.disp().window.restoreTasks(self.oldCategories, self.oldTasks)


class InitialState(BaseState):
    def init(self, version):
        self.version = version

        super(InitialState, self).init('i', 1)

        if self.version == _PROTOVERSION:
            self.ui = self.disp().window.createIPhoneProgressFrame()
            self.ui.Started()

        self.pack('i', version)

    def handleNewObject(self, accepted):
        if accepted:
            self.disp().log(_('Protocol version: %d'), self.version)
            self.setState(PasswordState)
        else:
            if self.version == 1:
                # Do not close the connection because it causes an error on the
                # device. It will do it itself.
                self.disp().window.notifyIPhoneProtocolFailed()
            else:
                self.disp().log(_('Rejected protocol version %d'), self.version)
                self.setState(InitialState, self.version - 1)

    def finished(self):
        pass


class PasswordState(BaseState):
    def init(self):
        super(PasswordState, self).init('20b', 1)

        self.hashData = ''.join([struct.pack('B', random.randint(0, 255)) for dummy in xrange(512)])
        self.pack('20b', self.hashData)

    def handleNewObject(self, hash): # pylint: disable=W0622
        local = hashlib.sha1()
        local.update(self.hashData + self.disp().settings.get('iphone', 'password').encode('UTF-8'))

        if hash == local.digest():
            self.disp().log(_('Hash OK.'))
            self.pack('i', 1)
            self.setState(DeviceNameState)
        else:
            self.disp().log(_('Hash KO.'))
            self.pack('i', 0)
            self.setState(PasswordState)

    def finished(self):
        pass


class DeviceNameState(BaseState):
    def init(self):
        super(DeviceNameState, self).init('s', 1)

    def handleNewObject(self, name):
        self.disp().log(_('Device name: %s'), name)
        self.deviceName = name
        self.ui.SetDeviceName(name)
        self.setState(GUIDState)


class GUIDState(BaseState):
    def init(self):
        if self.version >= 4:
            super(GUIDState, self).init('i', 1)
            self.pack('s', self.disp().window.taskFile.guid())
        else:
            super(GUIDState, self).init('z', 1)

    def handleNewObject(self, guid):
        self.disp().log(_('GUID: %s'), guid)

        if self.version >= 4:
            self.setState(TaskFileNameState)
        else:
            type_ = self.disp().window.getIPhoneSyncType(guid)

            self.pack('i', type_)

            if type_ == 0:
                self.setState(TwoWayState)
            elif type_ == 1:
                self.setState(FullFromDesktopState)
            elif type_ == 2:
                self.setState(FullFromDeviceState)

            # On cancel, the other end will close the connection

    def finished(self):
        pass


class TaskFileNameState(BaseState):
    def init(self):
        super(TaskFileNameState, self).init('i', 1)

        filename = self.disp().iocontroller.filename()
        if filename:
            filename = os.path.splitext(os.path.split(filename)[1])[0]
        self.disp().log(_('Sending file name: %s'), filename)
        self.pack('z', filename)

    def handleNewObject(self, response): # pylint: disable=W0613
        self.setState(TwoWayState if self.version < 5 else DayHoursState)
        
    def finished(self):
        pass


class DayHoursState(BaseState):
    def init(self):
        super(DayHoursState, self).init('i', 1)

        self.pack('ii',
                  self.disp().settings.getint('view', 'efforthourstart'),
                  self.disp().settings.getint('view', 'efforthourend'))

    def handleNewObject(self, response): # pylint: disable=W0613
        self.setState(TwoWayState)

    def finished(self):
        pass


class FullFromDesktopState(BaseState):
    def init(self):
        self.disp().log(_('Full from desktop.'))

        if self.version >= 4:
            allEfforts = self.disp().window.taskFile.efforts()

            if self.syncCompleted:
                self.tasks = list([task for task in self.disp().window.taskFile.tasks().allItemsSorted() if not task.isDeleted()])
                self.efforts = list([effort for effort in  allEfforts \
                                  if effort.task() is None or not effort.task().isDeleted()])
            else:
                self.tasks = list([task for task in self.disp().window.taskFile.tasks().allItemsSorted() if not (task.isDeleted() or task.completed())])
                self.efforts = list([effort for effort in allEfforts \
                                  if effort.task() is None or not (effort.task().isDeleted() or effort.task().completed())])
        else:
            self.tasks = filter(self.isTaskEligible, self.disp().window.taskFile.tasks()) # pylint: disable=W0141
        self.categories = list([cat for cat in self.disp().window.taskFile.categories().allItemsSorted() if not cat.isDeleted()])

        if self.version >= 4:
            self.pack('iii', len(self.categories), len(self.tasks), len(self.efforts))
            self.total = len(self.categories) + len(self.tasks) + len(self.efforts)
        else:
            self.pack('ii', len(self.categories), len(self.tasks))
            self.total = len(self.categories) + len(self.tasks)

        self.count = 0

        self.setState(FullFromDesktopCategoryState)


class FullFromDesktopCategoryState(BaseState):
    def init(self):
        super(FullFromDesktopCategoryState, self).init('i', len(self.categories))

        self.disp().log(_('%d categories'), len(self.categories))

        if self.categories:
            self.sendObject()

    def sendObject(self):
        if self.categories:
            category = self.categories.pop(0)
            self.disp().log(_('Send category %s'), category.id())
            self.pack('ssz', category.subject(), category.id(),
                      None if category.parent() is None else category.parent().id())

    def handleNewObject(self, code):
        self.disp().log(_('Response: %d'), code)
        self.count += 1
        self.ui.SetProgress(self.count, self.total)
        self.sendObject()

    def finished(self):
        self.setState(FullFromDesktopTaskState)


class FullFromDesktopTaskState(BaseState):
    def init(self):
        super(FullFromDesktopTaskState, self).init('i', len(self.tasks))

        self.disp().log(_('%d tasks'), len(self.tasks))

        if self.tasks:
            self.sendObject()

    def sendObject(self):
        if self.tasks:
            task = self.tasks.pop(0)
            self.disp().log(_('Send task %s'), task.id())
            if self.version < 4:
                self.pack('sssddd[s]',
                          task.subject(),
                          task.id(),
                          task.description(),
                          task.plannedStartDateTime().date(),
                          task.dueDateTime().date(),
                          task.completionDateTime().date(),
                          [category.id() for category in task.categories()])
            elif self.version < 5:
                self.pack('sssdddz[s]',
                          task.subject(),
                          task.id(),
                          task.description(),
                          task.plannedStartDateTime().date(),
                          task.dueDateTime().date(),
                          task.completionDateTime().date(),
                          task.parent().id() if task.parent() is not None else None,
                          [category.id() for category in task.categories()])
            else:
                hasRecurrence = task.recurrence() is not None and task.recurrence().unit != ''
                if hasRecurrence:
                    recPeriod = {'daily': 0, 'weekly': 1, 'monthly': 2, 'yearly': 3}[task.recurrence().unit]
                    recRepeat = task.recurrence().amount
                    recSameWeekday = task.recurrence().sameWeekday
                else:
                    recPeriod = 0
                    recRepeat = 0
                    recSameWeekday = 0

                self.pack('sssffffziiiii[s]',
                          task.subject(),
                          task.id(),
                          task.description(),
                          task.plannedStartDateTime(),
                          task.dueDateTime(),
                          task.completionDateTime(),
                          task.reminder(),
                          task.parent().id() if task.parent() is not None else None,
                          task.priority(),
                          hasRecurrence,
                          recPeriod,
                          recRepeat,
                          recSameWeekday,
                          [category.id() for category in task.categories()])

    def handleNewObject(self, code):
        self.disp().log(_('Response: %d'), code)
        self.count += 1
        self.ui.SetProgress(self.count, self.total)
        self.sendObject()

    def finished(self):
        if self.version >= 4:
            self.setState(FullFromDesktopEffortState)
        else:
            self.setState(SendGUIDState)


class FullFromDesktopEffortState(BaseState):
    def init(self):
        super(FullFromDesktopEffortState, self).init('i', len(self.efforts))

        self.disp().log(_('%d efforts'), len(self.efforts))

        if self.efforts:
            self.sendObject()

    def sendObject(self):
        if self.efforts:
            effort = self.efforts.pop(0)
            self.disp().log(_('Send effort %s'), effort.id())
            self.pack('ssztt',
                      effort.id(),
                      effort.subject(),
                      effort.task().id() if effort.task() is not None else None,
                      effort.getStart(),
                      effort.getStop())

    def handleNewObject(self, code): # pylint: disable=W0613
        self.count += 1
        self.ui.SetProgress(self.count, self.total)
        self.sendObject()

    def finished(self):
        if self.version < 5:
            self.setState(SendGUIDState)
        else:
            self.disp().log(_('Finished.'))
            self.disp().close_when_done()
            self.ui.Finished()

    def handleClose(self):
        if self.version < 5:
            super(FullFromDesktopEffortState, self).handleClose()


class FullFromDeviceState(BaseState):
    def init(self):
        self.disp().window.clearTasks()

        super(FullFromDeviceState, self).init('ii', 1)

    def handleNewObject(self, (categoryCount, taskCount)):
        self.categoryCount = categoryCount
        self.taskCount = taskCount

        self.total = categoryCount + taskCount
        self.count = 0

        self.setState(FullFromDeviceCategoryState)

    def finished(self):
        pass


class FullFromDeviceCategoryState(BaseState):
    def init(self):
        self.categoryMap = {}

        super(FullFromDeviceCategoryState, self).init('s' if self.version < 3 else 'sz', self.categoryCount)

    def handleNewObject(self, args):
        if self.version < 3:
            name = args
            parentId = None
        else:
            name, parentId = args

        if parentId is None:
            category = Category(name)
        else:
            category = self.categoryMap[parentId].newChild(name)

        self.disp().window.addIPhoneCategory(category)

        self.pack('s', category.id())
        self.categoryMap[category.id()] = category

        self.count += 1
        self.ui.SetProgress(self.count, self.total)

    def finished(self):
        self.setState(FullFromDeviceTaskState)


class FullFromDeviceTaskState(BaseState):
    def init(self):
        super(FullFromDeviceTaskState, self).init('ssddd[s]', self.taskCount)

    def handleNewObject(self, (subject, description, startDate, dueDate, completionDate, categories)):
        task = Task(subject=subject, description=description, 
                    plannedStartDateTime=DateTime(startDate.year, startDate.month, startDate.day),
                    dueDateTime=DateTime(dueDate.year, dueDate.month, dueDate.day), 
                    completionDateTime=DateTime(completionDate.year, completionDate.month, completionDate.day))

        self.disp().window.addIPhoneTask(task, [self.categoryMap[id_] for id_ in categories])

        self.count += 1
        self.ui.SetProgress(self.count, self.total)

        self.pack('s', task.id())

    def finished(self):
        self.setState(SendGUIDState)


class TwoWayState(BaseState):
    def init(self):
        self.categoryMap = dict([(category.id(), category) for category in self.disp().window.taskFile.categories()])
        self.taskMap = dict([(task.id(), task) for task in self.disp().window.taskFile.tasks()])
        self.effortMap = dict([(effort.id(), effort) for effort in self.disp().window.taskFile.efforts()])

        if self.version < 3:
            super(TwoWayState, self).init('iiii', 1)
        elif self.version < 4:
            super(TwoWayState, self).init('iiiiii', 1)
        else:
            super(TwoWayState, self).init('iiiiiiiii', 1)

    def handleNewObject(self, args):
        if self.version < 3:
            (self.newCategoriesCount,
             self.newTasksCount,
             self.deletedTasksCount,
             self.modifiedTasksCount) = args
        elif self.version < 4:
            (self.newCategoriesCount,
             self.newTasksCount,
             self.deletedTasksCount,
             self.modifiedTasksCount,
             self.deletedCategoriesCount,
             self.modifiedCategoriesCount) = args
        else:
            (self.newCategoriesCount,
             self.newTasksCount,
             self.deletedTasksCount,
             self.modifiedTasksCount,
             self.deletedCategoriesCount,
             self.modifiedCategoriesCount,
             self.newEffortsCount,
             self.modifiedEffortsCount,
             self.deletedEffortsCount) = args

            self.disp().log(_('%d new categories'), self.newCategoriesCount)
            self.disp().log(_('%d new tasks'), self.newTasksCount)
            self.disp().log(_('%d new efforts'), self.newEffortsCount)
            self.disp().log(_('%d modified categories'), self.modifiedCategoriesCount)
            self.disp().log(_('%d modified tasks'), self.modifiedTasksCount)
            self.disp().log(_('%d modified efforts'), self.modifiedEffortsCount)
            self.disp().log(_('%d deleted categories'), self.deletedCategoriesCount)
            self.disp().log(_('%d deleted tasks'), self.deletedTasksCount)
            self.disp().log(_('%d deleted efforts'), self.deletedEffortsCount)

        self.setState(TwoWayNewCategoriesState)


class TwoWayNewCategoriesState(BaseState):
    def init(self):
        super(TwoWayNewCategoriesState, self).init(('s' if self.version < 3 else 'sz'), self.newCategoriesCount)

    def handleNewObject(self, args):
        if self.version < 3:
            name = args
            parentId = None
        else:
            name, parentId = args
            self.disp().log(_('New category (parent: %s)'), parentId)

        if parentId is None or not self.categoryMap.has_key(parentId):
            category = Category(name)
        else:
            category = self.categoryMap[parentId].newChild(name)

        self.disp().window.addIPhoneCategory(category)

        self.categoryMap[category.id()] = category
        self.pack('s', category.id())

    def finished(self):
        if self.version < 3:
            self.setState(TwoWayNewTasksState)
        else:
            self.setState(TwoWayDeletedCategoriesState)


class TwoWayDeletedCategoriesState(BaseState):
    def init(self):
        super(TwoWayDeletedCategoriesState, self).init('s', self.deletedCategoriesCount)

    def handleNewObject(self, catId):
        try:
            category = self.categoryMap.pop(catId)
        except KeyError:
            # Deleted on desktop
            if self.version >= 5:
                self.pack('s', '')
        else:
            self.disp().log(_('Delete category %s'), category.id())
            if self.version >= 5:
                self.pack('s', category.id())
            self.disp().window.removeIPhoneCategory(category)

    def finished(self):
        self.setState(TwoWayModifiedCategoriesState)


class TwoWayModifiedCategoriesState(BaseState):
    def init(self):
        super(TwoWayModifiedCategoriesState, self).init('ss', self.modifiedCategoriesCount)

    def handleNewObject(self, (name, catId)):
        try:
            category = self.categoryMap[catId]
        except KeyError:
            if self.version >= 5:
                self.pack('s', '')
        else:
            self.disp().log(_('Modify category %s'), category.id())
            self.disp().window.modifyIPhoneCategory(category, name)

            if self.version >= 5:
                self.pack('s', category.id())

    def finished(self):
        if self.version < 4:
            self.setState(TwoWayNewTasksState)
        elif self.version < 5:
            self.setState(TwoWayNewTasksState4)
        else:
            self.setState(TwoWayNewTasksState5)


class TwoWayNewTasksState(BaseState):
    def init(self):
        super(TwoWayNewTasksState, self).init('ssddd[s]', self.newTasksCount)

    def handleNewObject(self, (subject, description, startDate, dueDate, completionDate, categories)):
        task = Task(subject=subject, description=description, 
                    plannedStartDateTime=DateTime(startDate.year, startDate.month, startDate.day),
                    dueDateTime=DateTime(dueDate.year, dueDate.month, dueDate.day), 
                    completionDateTime=DateTime(completionDate.year, completionDate.month, completionDate.day))

        self.disp().window.addIPhoneTask(task, [self.categoryMap[catId] for catId in categories \
                                                    if self.categoryMap.has_key(catId)])
        self.disp().log(_('New task %s'), task.id())

        self.taskMap[task.id()] = task
        self.pack('s', task.id())

    def finished(self):
        self.setState(TwoWayDeletedTasksState)


class TwoWayNewTasksState4(BaseState):
    def init(self):
        super(TwoWayNewTasksState4, self).init('ssddfz[s]', self.newTasksCount)

    def handleNewObject(self, (subject, description, plannedStartDate, dueDate, completionDateTime, parentId, categories)):
        parent = self.taskMap[parentId] if parentId and self.taskMap.has_key(parentId) else None

        if self.version < 5:
            plannedStartDateTime = DateTime() if plannedStartDate == Date() else \
                DateTime(year=plannedStartDate.year, month=plannedStartDate.month,
                         day=plannedStartDate.day, hour=self.disp().settings.getint('view', 'efforthourstart'))

            dueDateTime = DateTime() if dueDate == Date() else \
                DateTime(year=dueDate.year, month=dueDate.month, day=dueDate.day,
                         hour=self.disp().settings.getint('view', 'efforthourend'))

        task = Task(subject=subject, description=description, 
                    plannedStartDateTime=plannedStartDateTime,
                    dueDateTime=dueDateTime, 
                    completionDateTime=completionDateTime, 
                    parent=parent)

        self.disp().window.addIPhoneTask(task, [self.categoryMap[catId] for catId in categories \
                                                    if self.categoryMap.has_key(catId)])
        self.disp().log(_('New task %s'), task.id())

        self.taskMap[task.id()] = task
        self.pack('s', task.id())

    def finished(self):
        self.setState(TwoWayDeletedTasksState)


class TwoWayNewTasksState5(BaseState):
    def init(self):
        super(TwoWayNewTasksState5, self).init('ssffffiiiiiz[s]', self.newTasksCount)

    def handleNewObject(self, (subject, description, plannedStartDateTime, dueDateTime, completionDateTime,
                               reminderDateTime, priority, hasRecurrence, recPeriod, recRepeat,
                               recSameWeekday, parentId, categories)):
        parent = self.taskMap[parentId] if parentId else None

        recurrence = None
        if hasRecurrence:
            recurrence = Recurrence(unit={0: 'daily', 1: 'weekly', 2: 'monthly', 3: 'yearly'}[recPeriod],
                                    amount=recRepeat, sameWeekday=recSameWeekday)

        task = Task(subject=subject, description=description, 
                    plannedStartDateTime=plannedStartDateTime,
                    dueDateTime=dueDateTime, 
                    completionDateTime=completionDateTime, 
                    parent=parent,
                    recurrence=recurrence,
                    priority=priority)

        # Don't start a timer from this thread...
        wx.CallAfter(task.setReminder, reminderDateTime)

        self.disp().window.addIPhoneTask(task, [self.categoryMap[catId] for catId in categories \
                                                    if self.categoryMap.has_key(catId)])
        self.disp().log(_('New task %s'), task.id())

        self.taskMap[task.id()] = task
        self.pack('s', task.id())

    def finished(self):
        self.setState(TwoWayDeletedTasksState)


class TwoWayDeletedTasksState(BaseState):
    def init(self):
        super(TwoWayDeletedTasksState, self).init('s', self.deletedTasksCount)

    def handleNewObject(self, taskId):
        try:
            task = self.taskMap.pop(taskId)
        except KeyError:
            if self.version >= 5:
                self.pack('s', '')
        else:
            self.disp().log(_('Delete task %s'), task.id())
            if self.version >= 5:
                self.pack('s', task.id())
            self.disp().window.removeIPhoneTask(task)

    def finished(self):
        self.setState(TwoWayModifiedTasks)


class TwoWayModifiedTasks(BaseState):
    def init(self):
        if self.version < 2:
            super(TwoWayModifiedTasks, self).init('sssddd', self.modifiedTasksCount)
        elif self.version < 5:
            super(TwoWayModifiedTasks, self).init('sssddd[s]', self.modifiedTasksCount)
        else:
            super(TwoWayModifiedTasks, self).init('sssffffiiiii[s]', self.modifiedTasksCount)

    def handleNewObject(self, args):
        reminderDateTime = None
        recurrence = None
        priority = 0

        if self.version < 2:
            subject, taskId, description, plannedStartDate, dueDate, completionDate = args
            categories = None
        elif self.version < 5:
            subject, taskId, description, plannedStartDate, dueDate, completionDate, categories = args
            categories = set([self.categoryMap[catId] for catId in categories])
        else:
            (subject, taskId, description, plannedStartDate, dueDate, completionDate, reminderDateTime,
             priority, hasRecurrence, recPeriod, recRepeat, recSameWeekday, categories) = args
            categories = set([self.categoryMap[catId] for catId in categories if catId in self.categoryMap])

            if hasRecurrence:
                recurrence = Recurrence(unit={0: 'daily', 1: 'weekly', 2: 'monthly', 3: 'yearly'}[recPeriod],
                                        amount=recRepeat, sameWeekday=recSameWeekday)

        if self.version < 5:
            plannedStartDateTime = DateTime(plannedStartDate.year, plannedStartDate.month, plannedStartDate.day,
                self.disp().settings.getint('view', 'efforthourstart')) if plannedStartDate != Date() else DateTime()
            dueDateTime = DateTime(dueDate.year, dueDate.month, dueDate.day,
                self.disp().settings.getint('view', 'efforthourend')) if dueDate != Date() else DateTime()
            completionDateTime = DateTime(completionDate.year, completionDate.month, 
                completionDate.day) if completionDate != Date() else DateTime()
        else:
            plannedStartDateTime = plannedStartDate
            dueDateTime = dueDate
            completionDateTime = completionDate

        try:
            task = self.taskMap[taskId]
        except KeyError:
            if self.version >= 5:
                self.pack('s', '')
        else:
            self.disp().log(_('Modify task %s'), task.id())
            self.disp().window.modifyIPhoneTask(task, subject, description, 
                                                plannedStartDateTime, dueDateTime, 
                                                completionDateTime, reminderDateTime,
                                                recurrence, priority, categories)
            if self.version >= 5:
                self.pack('s', task.id())

    def finished(self):
        self.disp().log(_('End of task synchronization.'))
        if self.version < 4:
            self.setState(FullFromDesktopState)
        else:
            self.setState(TwoWayNewEffortsState)


class TwoWayNewEffortsState(BaseState):
    def init(self):
        super(TwoWayNewEffortsState, self).init('sztt', self.newEffortsCount)

    def handleNewObject(self, (subject, taskId, started, ended)):
        task = None
        if taskId is not None:
            try:
                task = self.taskMap[taskId]
            except KeyError:
                self.disp().log(_('Could not find task %s for effort.'), taskId)

        effort = Effort(task, started, ended, subject=subject)
        self.disp().log(_('New effort %s'), effort.id())
        self.disp().window.addIPhoneEffort(task, effort)

        self.pack('s', effort.id())

        self.effortMap[effort.id()] = effort

    def finished(self):
        self.setState(TwoWayModifiedEffortsState)


class TwoWayModifiedEffortsState(BaseState):
    def init(self):
        super(TwoWayModifiedEffortsState, self).init('sstt', self.modifiedEffortsCount)

    def handleNewObject(self, (id_, subject, started, ended)):
        # Actually, the taskId cannot be modified on the device, which saves
        # us some headaches.

        try:
            effort = self.effortMap[id_]
        except KeyError:
            if self.version >= 5:
                self.pack('s', '')
        else:
            self.disp().log(_('Modify effort %s'), effort.id())
            self.disp().window.modifyIPhoneEffort(effort, subject, started, ended)
            if self.version >= 5:
                self.pack('s', effort.id())

    def finished(self):
        # Efforts cannot be deleted on the iPhone yet.
        self.setState(FullFromDesktopState)


class SendGUIDState(BaseState):
    def init(self):
        super(SendGUIDState, self).init('i', 1)

        self.disp().log(_('Sending GUID: %s'), self.disp().window.taskFile.guid())
        self.pack('s', self.disp().window.taskFile.guid())

    def handleNewObject(self, code):
        pass

    def finished(self):
        self.disp().log(_('Finished.'))
        self.disp().close_when_done()
        self.ui.Finished()

    def handleClose(self):
        pass



'''
Task Coach - Your friendly task manager
Copyright (C) 2011 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import wx # For ArtProvider

from taskcoachlib.changes import ChangeMonitor
from taskcoachlib.domain.note import NoteOwner
from taskcoachlib.domain.attachment import AttachmentOwner
from taskcoachlib.domain.base import CompositeObject
from taskcoachlib.domain.task import Task
from taskcoachlib.domain.note import Note
from taskcoachlib.domain.category import Category
from taskcoachlib.notify import AbstractNotifier
from taskcoachlib.i18n import _


class ChangeSynchronizer(object):
    def __init__(self, monitor, allChanges):
        self._monitor = monitor
        self._allChanges = allChanges

    @staticmethod
    def allObjects(theList):
        result = list()
        for obj in theList:
            result.append(obj)
            if isinstance(obj, CompositeObject):
                result.extend(ChangeSynchronizer.allObjects(obj.children()))
            if isinstance(obj, NoteOwner):
                result.extend(ChangeSynchronizer.allObjects(obj.notes()))
            if isinstance(obj, AttachmentOwner):
                result.extend(ChangeSynchronizer.allObjects(obj.attachments()))
            if isinstance(obj, Task):
                result.extend(obj.efforts())
        return result

    def sync(self, lists):
        self.diskChanges = ChangeMonitor()
        self.conflictChanges = ChangeMonitor()
        self.notifier = AbstractNotifier.getSimple()

        self.memMap = dict()
        self.memOwnerMap = dict()
        self.diskMap = dict()
        self.diskOwnerMap = dict()

        for devGUID, changes in self._allChanges.items():
            if devGUID == self._monitor.guid():
                self.diskChanges = changes
                break
        self._allChanges[self._monitor.guid()] = self._monitor

        for memList, diskList in lists:
            self.mergeObjects(memList, diskList)

        # Cleanup monitor
        self._monitor.empty()
        for memList, diskList in lists:
            for obj in self.allObjects(memList.rootItems()):
                self._monitor.resetChanges(obj)

        # Merge conflict changes
        for devGUID, changes in self._allChanges.items():
            if devGUID != self._monitor.guid():
                changes.merge(self.conflictChanges)

    def notify(self, message):
        self.notifier.notify(_('Task Coach'), message,
                             wx.ArtProvider.GetBitmap('taskcoach', size=wx.Size(32, 32)))

    def mergeObjects(self, memList, diskList):
        # Map id to object
        def addIds(objects, idMap, ownerMap, owner=None):
            for obj in objects:
                idMap[obj.id()] = obj
                if owner is not None:
                    ownerMap[obj.id()] = owner
                if isinstance(obj, CompositeObject):
                    addIds(obj.children(), idMap, ownerMap)
                if isinstance(obj, NoteOwner):
                    addIds(obj.notes(), idMap, ownerMap, obj)
                if isinstance(obj, AttachmentOwner):
                    addIds(obj.attachments(), idMap, ownerMap, obj)
                if isinstance(obj, Task):
                    addIds(obj.efforts(), idMap, ownerMap)
        addIds(memList, self.memMap, self.memOwnerMap)
        addIds(diskList, self.diskMap, self.diskOwnerMap)

        self.mergeCompositeObjects(memList, diskList)
        self.mergeOwnedObjectsFromDisk(diskList)
        self.reparentObjects(memList, diskList)
        self.deletedObjects(memList)
        self.deletedOwnedObjects(memList)
        self.applyChanges(memList)

    def mergeCompositeObjects(self, memList, diskList):
        # First pass: new composite objects on disk. Don't handle
        # other (notes belonging to object, attachments, efforts) yet
        # because they may belong to a category. This assumes that
        # categories are the first domain class handled.

        for diskObject in diskList.allItemsSorted():
            memChanges = self._monitor.getChanges(diskObject)
            deleted = memChanges is not None and '__del__' in memChanges
            diskChanges = self.diskChanges.getChanges(diskObject)
            if deleted and diskChanges is not None and '__del__' not in diskChanges and len(diskChanges) > 0:
                # "undelete" it
                memChanges.remove('__del__')
                deleted = False

            if diskObject.id() not in self.memMap and not deleted:
                if isinstance(diskObject, CompositeObject):
                    # New children will be handled later. This assumes
                    # that the parent() is not changed when removing a
                    # child.
                    for child in diskObject.children():
                        diskObject.removeChild(child)
                    parent = diskObject.parent()
                    if parent is not None and parent.id() in self.memMap:
                        parent = self.memMap[parent.id()]
                        parent.addChild(diskObject)
                        diskObject.setParent(parent)
                    elif parent is not None:
                        # Parent deleted from memory; the task will be
                        # top-level.
                        diskObject.setParent(None)
                        self.conflictChanges.addChange(diskObject, '__parent__')
                        self.notify(_('"%s" became top-level because its parent was locally deleted.') %
                                    diskObject.subject())
                memList.append(diskObject)
                self.memMap[diskObject.id()] = diskObject

    def mergeOwnedObjectsFromDisk(self, diskList):
        # Second pass: 'owned' objects (notes and attachments
        # currently) new on disk, and efforts.

        for obj in diskList.allItemsSorted():
            if isinstance(obj, NoteOwner):
                self._handleNewOwnedObjectsOnDisk(obj.notes())
            if isinstance(obj, AttachmentOwner):
                self._handleNewOwnedObjectsOnDisk(obj.attachments())
            if isinstance(obj, Task):
                self._handleNewEffortsOnDisk(obj.efforts())

    def _handleNewOwnedObjectsOnDisk(self, diskObjects):
        for diskObject in diskObjects:
            className = diskObject.__class__.__name__
            if className.endswith('Attachment'):
                className = 'Attachment'

            if isinstance(diskObject, CompositeObject):
                children = diskObject.children()[:]

            memChanges = self._monitor.getChanges(diskObject)
            deleted = memChanges is not None and '__del__' in memChanges

            if diskObject.id() not in self.memMap and not deleted:
                addObject = True

                if isinstance(diskObject, CompositeObject):
                    for child in diskObject.children():
                        diskObject.removeChild(child)
                    parent = diskObject.parent()
                    if parent is not None and parent.id() in self.memMap:
                        parent = self.memMap[parent.id()]
                        parent.addChild(diskObject)
                        diskObject.setParent(parent)
                    elif parent is not None:
                        # Parent deleted from memory; the object
                        # becomes top-level but its owner stays
                        # the same.
                        diskObject.setParent(None)
                        while parent.parent() is not None:
                            parent = parent.parent()
                        diskOwner = self.diskOwnerMap[parent.id()]
                        if diskOwner.id() in self.memMap:
                            memOwner = self.memMap[diskOwner.id()]
                            getattr(memOwner, 'add%s' % className)(diskObject)
                            self.conflictChanges.addChange(diskObject, '__owner__')
                            self.memOwnerMap[diskObject.id()] = memOwner
                        self.notify(_('"%s" became top-level because its parent was locally deleted.') %
                                    diskObject.subject())
                    else:
                        diskOwner = self.diskOwnerMap[diskObject.id()]
                        if diskOwner.id() in self.memMap:
                            memOwner = self.memMap[diskOwner.id()]
                            getattr(memOwner, 'add%s' % className)(diskObject)
                            self.memOwnerMap[diskObject.id()] = memOwner
                        else:
                            # Owner deleted. Just forget it.
                            self.conflictChanges.addChange(diskObject, '__del__')
                            addObject = False
                else:
                    diskOwner = self.diskOwnerMap[diskObject.id()]
                    if diskOwner.id() in self.memMap:
                        memOwner = self.memMap[diskOwner.id()]
                        getattr(memOwner, 'add%s' % className)(diskObject)
                        self.memOwnerMap[diskObject.id()] = memOwner
                    else:
                        # Forget it again...
                        self.conflictChanges.addChange(diskObject, '__del__')
                        addObject = False

                if addObject:
                    self.memMap[diskObject.id()] = diskObject

            if diskObject.id() in self.memMap:
                if isinstance(diskObject, CompositeObject):
                    self._handleNewOwnedObjectsOnDisk(children)
                if isinstance(diskObject, NoteOwner):
                    self._handleNewOwnedObjectsOnDisk(diskObject.notes())
                if isinstance(diskObject, AttachmentOwner):
                    self._handleNewOwnedObjectsOnDisk(diskObject.attachments())

    def _handleNewEffortsOnDisk(self, diskEfforts):
        for diskEffort in diskEfforts:
            memChanges = self._monitor.getChanges(diskEffort)
            deleted = memChanges is not None and '__del__' in memChanges
            if diskEffort.id() not in self.memMap and not deleted:
                diskTask = diskEffort.parent()
                if diskTask.id() in self.memMap:
                    memTask = self.memMap[diskTask.id()]
                    diskEffort.setTask(memTask)
                    self.memMap[diskEffort.id()] = diskEffort
                else:
                    # Task deleted; forget it.
                    self.conflictChanges.addChange(diskEffort, '__del__')

    def reparentObjects(self, memList, diskList):
        # Third pass: objects reparented on disk.

        for diskObject in self.allObjects(diskList):
            diskChanges = self.diskChanges.getChanges(diskObject)
            if diskChanges is not None and '__parent__' in diskChanges:
                memChanges = self._monitor.getChanges(diskObject)
                memObject = self.memMap[diskObject.id()]
                memList.remove(memObject)

                # Note: no conflict resolution for this one,
                # it would be a bit tricky... Instead, the
                # disk version wins.

                def sameParents(a, b):
                    if a is None and b is None:
                        return True
                    elif a is None or b is None:
                        return False
                    return a.id() == b.id()

                parentConflict = False
                if memChanges is not None and '__parent__' in memChanges:
                    if not sameParents(memObject.parent(), diskObject.parent()):
                        parentConflict = True

                if memObject.parent() is not None:
                    memObject.parent().removeChild(memObject)

                if parentConflict:
                    diskParent = diskObject.parent()

                    if diskParent is None:
                        memObject.setParent(None)
                    else:
                        if diskParent.id() in self.memMap:
                            memParent = self.memMap[diskParent.id()]
                            memParent.addChild(memObject)
                            memObject.setParent(memParent)
                        else:
                            # New parent deleted from memory...
                            memObject.setParent(None)
                            self.conflictChanges.addChange(memObject, '__parent__')
                else:
                    diskParent = diskObject.parent()
                    if diskParent is None:
                        memObject.setParent(None)
                    else:
                        if diskParent.id() in self.memMap:
                            memParent = self.memMap[diskParent.id()]
                            memParent.addChild(memObject)
                            memObject.setParent(memParent)
                        else:
                            memObject.setParent(None)
                            self.conflictChanges.addChange(memObject, '__parent__')

                memList.append(memObject)

    def deletedObjects(self, memList):
        # Fourth pass: objects deleted from disk

        for memObject in memList.allItemsSorted():
            diskChanges = self.diskChanges.getChanges(memObject)
            memChanges = self._monitor.getChanges(memObject)

            if diskChanges is not None and '__del__' in diskChanges:
                if (memChanges is None or '__del__' in memChanges or len(memChanges) == 0):
                    memList.remove(memObject)
                    del self.memMap[memObject.id()]
                    if memObject.id() in self.memOwnerMap:
                        del self.memOwnerMap[memObject.id()]
                else:
                    # If there are local changes they win over deletion.
                    self.diskMap[memObject.id()] = memObject
                    self.diskChanges.resetChanges(memObject)

    def deletedOwnedObjects(self, memList):
        for obj in memList.allItemsSorted():
            if isinstance(obj, NoteOwner):
                self._handleOwnedObjectsRemovedFromDisk(obj.notes())
            if isinstance(obj, AttachmentOwner):
                self._handleOwnedObjectsRemovedFromDisk(obj.attachments())
            if isinstance(obj, Task):
                self._handleEffortsRemovedFromDisk(obj.efforts())

    def _handleOwnedObjectsRemovedFromDisk(self, memObjects):
        for memObject in memObjects:
            className = memObject.__class__.__name__
            if className.endswith('Attachment'):
                className = 'Attachment'

            if isinstance(memObject, CompositeObject):
                self._handleOwnedObjectsRemovedFromDisk(memObject.children())
            if isinstance(memObject, NoteOwner):
                self._handleOwnedObjectsRemovedFromDisk(memObject.notes())
            if isinstance(memObject, AttachmentOwner):
                self._handleOwnedObjectsRemovedFromDisk(memObject.attachments())

            diskChanges = self.diskChanges.getChanges(memObject)
            if diskChanges is not None and '__del__' in diskChanges:
                # Same remark as above
                if isinstance(memObject, CompositeObject):
                    if memObject.parent() is None:
                        getattr(self.memOwnerMap[memObject.id()], 'remove%s' % className)(memObject)
                    else:
                        self.memMap[memObject.parent().id()].removeChild(memObject)
                else:
                    getattr(self.memOwnerMap[memObject.id()], 'remove%s' % className)(memObject)
                del self.memMap[memObject.id()]

    def _handleEffortsRemovedFromDisk(self, memEfforts):
        for memEffort in memEfforts:
            diskChanges = self.diskChanges.getChanges(memEffort)

            if diskChanges is not None and '__del__' in diskChanges:
                # Same remark as above
                self.memMap[memEffort.parent().id()].removeEffort(memEffort)
                del self.memMap[memEffort.id()]

    def applyChanges(self, memList):
        # Final: apply disk changes

        for memObject in self.allObjects(memList.rootItems()):
            diskChanges = self.diskChanges.getChanges(memObject)
            if diskChanges:
                memChanges = self._monitor.getChanges(memObject)
                diskObject = self.diskMap[memObject.id()]

                conflicts = []

                for changeName in diskChanges:
                    if changeName == '__parent__':
                        pass # Already handled
                    elif changeName.startswith('__add_category:'):
                        categoryId = changeName[15:]
                        if categoryId not in self.memMap:
                            # Mmmh, deleted...
                            conflicts.append(changeName)
                            self.conflictChanges.addChange(memObject, '__del' + changeName[5:])
                        else:
                            if memChanges is not None and \
                               '__del' + changeName[5:] in memChanges:
                                conflicts.append(changeName)
                                self.conflictChanges.addChange(memObject, '__del' + changeName[5:])
                            else:
                                # Aaaaah finally
                                theCategory = self.memMap[categoryId]
                                memObject.addCategory(theCategory)
                                theCategory.addCategorizable(memObject)
                    elif changeName.startswith('__del_category:'):
                        categoryId = changeName[15:]
                        if categoryId in self.memMap:
                            if memChanges is not None and \
                               '__add' + changeName[5:] in memChanges:
                                conflicts.append(changeName)
                                self.conflictChanges.addChange(memObject, '__add' + changeName[5:])
                            else:
                                theCategory = self.memMap[categoryId]
                                memObject.removeCategory(theCategory)
                                theCategory.removeCategorizable(memObject)
                    elif changeName == '__prerequisites__':
                        diskPrereqs = set([self.memMap[obj.id()] for obj in diskObject.prerequisites()])
                        memPrereqs = set(memObject.prerequisites())
                        if memChanges is not None and \
                           '__prerequisites__' in memChanges and \
                           memPrereqs != diskPrereqs:
                            conflicts.append('__prerequisites__')
                            self.conflictChanges.addChange(memObject, '__prerequisites__')
                        else:
                            memObject.setPrerequisites(diskPrereqs)
                    elif changeName == '__task__':
                        # Effort changed task
                        if memChanges is not None and \
                           '__task__' in memChanges and \
                           memObject.parent().id() != diskObject.parent().id():
                            conflicts.append('__task__')
                            self.conflictChanges.addChange(memObject, '__task__')
                        else:
                            memObject.setTask(self.memMap[diskObject.parent().id()])
                    elif changeName == '__owner__':
                        # This happens after a conflict
                        if memChanges is not None and \
                           '__owner__' in memChanges and \
                           self.memOwnerMap[memObject.id()].id() != self.diskOwnerMap[diskObject.id()].id():
                            # Yet another conflict... Memory wins
                            conflicts.append('__owner__')
                            self.conflictChanges.addChange(memObject, '__owner__')
                        else:
                            className = memObject.__class__.__name__
                            if className.endsWith('Attachment'):
                                className = 'Attachment'
                            oldOwner = self.memOwnerMap[memObject.id()]
                            newOwner = self.memOwnerMap[diskObject.id()]
                            getattr(oldOwner, 'remove%s' % className)(memObject)
                            getattr(newOwner, 'add%s' % className)(memObject)
                    elif changeName == 'appearance':
                        attrNames = ['foregroundColor', 'backgroundColor', 'font', 'icon', 'selectedIcon']
                        if memChanges is not None and \
                           'appearance' in memChanges:
                            for attrName in attrNames:
                                if getattr(memObject, attrName)() != getattr(diskObject, attrName)():
                                    conflicts.append(attrName)
                            self.conflictChanges.addChange(memObject, 'appearance')
                        else:
                            for attrName in attrNames:
                                getattr(memObject, 'set' + attrName[0].upper() + attrName[1:])(getattr(diskObject, attrName)())
                    elif changeName == 'expandedContexts':
                        # Note: no conflict resolution for this one.
                        memObject.expand(diskObject.isExpanded())
                    else:
                        if changeName in ['start', 'stop']:
                            getterName = 'get' + changeName[0].upper() + changeName[1:]
                        else:
                            getterName = changeName
                        if memChanges is not None and \
                               changeName in memChanges and \
                               getattr(memObject, getterName)() != getattr(diskObject, getterName)():
                            conflicts.append(changeName)
                            self.conflictChanges.addChange(memObject, changeName)
                        else:
                            getattr(memObject, 'set' + changeName[0].upper() + changeName[1:])(getattr(diskObject, getterName)())

                    if conflicts:
                        self.notify(_('Conflicts detected for "%s".\nThe local version was used.') % memObject.subject())



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from taskcoachlib import gui, config, persistence, command, patterns, render, \
    operating_system
from taskcoachlib.domain import task, date, effort, category, attachment
from taskcoachlib.i18n import _
import locale
import os
import test
import wx
import weakref


class TaskViewerUnderTest(gui.viewer.task.TaskViewer):  # pylint: disable=W0223
    def __init__(self, *args, **kwargs):
        super(TaskViewerUnderTest, self).__init__(*args, **kwargs)
        self.events = []
        self.events_deprecated = []
        
    def onAttributeChanged(self, newValue, sender):
        super(TaskViewerUnderTest, self).onAttributeChanged(newValue, sender)
        self.events.append((newValue, sender))
    
    def onAttributeChanged_Deprecated(self, event):
        super(TaskViewerUnderTest, self).onAttributeChanged_Deprecated(event)
        self.events_deprecated.append(event)
        

class TaskViewerTestCase(test.wxTestCase):
    treeMode = 'Subclass responsibility'
    
    def setUp(self):
        super(TaskViewerTestCase, self).setUp()
        task.Task.settings = self.settings = config.Settings(load=False)
        self.task = task.Task(subject='task', plannedStartDateTime=date.Now())
        self.child = task.Task(subject='child', plannedStartDateTime=date.Now())
        self.child.setParent(self.task)
        self.taskFile = persistence.TaskFile()
        self.taskList = self.taskFile.tasks()
        self.parentFrame = wx.Frame(self.frame, wx.ID_ANY, '')
        self.viewer = TaskViewerUnderTest(self.parentFrame, self.taskFile, 
                                          self.settings)
        self.viewer.sortBy('subject')
        self.viewer.setSortOrderAscending()
        self.viewer.setSortByTaskStatusFirst(True)
        self.settings.setboolean(self.viewer.settingsSection(), 'treemode', 
                                 self.treeMode)
        self.newColor = (100, 200, 100, 255)
        attachment.Attachment.attdir = os.getcwd()
        if not operating_system.isGTK():
            self.originalLocale = locale.getlocale(locale.LC_ALL)
            tmpLocale = os.environ['LC_ALL'] if 'LC_ALL' in os.environ else ('en_US' if operating_system.isMac() else '')
            locale.setlocale(locale.LC_ALL, tmpLocale)

    def tearDown(self):
        super(TaskViewerTestCase, self).tearDown()
        if not operating_system.isGTK():
            locale.setlocale(locale.LC_ALL, self.originalLocale)
        attachment.Attachment.attdir = None
        self.taskFile.close()
        self.taskFile.stop()

        for name in os.listdir('.'):
            if os.path.isdir(name) and name.endswith('_attachments'):
                os.rmdir(name)  # pragma: no cover
        if os.path.isfile('test.mail'):
            os.remove('test.mail')

        if self.viewer:
            self.viewer.detach()
            self.viewer = None

        if self.parentFrame:
            self.parentFrame.Close()
            wx.Yield()

    def assertItems(self, *tasks):
        self.viewer.expandAll()  # pylint: disable=E1101
        self.assertEqual(self.viewer.size(), len(tasks))
        for index, eachTask in enumerate(tasks):
            self.assertItem(index, eachTask)
            
    def assertItem(self, index, aTask):
        if type(aTask) == type((), ):
            aTask, nrChildren = aTask
        else:
            nrChildren = 0
        subject = aTask.subject(recursive=not self.viewer.isTreeViewer())
        treeItem = self.viewer.widget.GetItemChildren(recursively=True)[index]
        self.assertEqual(subject, self.viewer.widget.GetItemText(treeItem))
        self.assertEqual(nrChildren, 
            self.viewer.widget.GetChildrenCount(treeItem, recursively=False))

    def firstItem(self):
        widget = self.viewer.widget
        return widget.GetFirstChild(widget.GetRootItem())[0]

    def getItemText(self, row, column):
        assert row == 0
        return self.viewer.widget.GetItemText(self.firstItem(), column)
                                             
    def getFirstItemTextColor(self):
        return self.viewer.widget.GetItemTextColour(self.firstItem())
    
    def getFirstItemBackgroundColor(self):
        return self.viewer.widget.GetItemBackgroundColour(self.firstItem())

    def getFirstItemFont(self):
        return self.viewer.widget.GetItemFont(self.firstItem())
    
    def getFirstItemIcon(self, column=0):
        return self.viewer.widget.GetItemImage(self.firstItem(), column=column)

    def showColumn(self, columnName, show=True):
        self.viewer.showColumnByName(columnName, show)

    def setColor(self, setting):
        self.settings.settuple('fgcolor', setting, self.newColor)       
         
    def assertColor(self, expectedColor=None):
        expectedColor = expectedColor or wx.Colour(*self.newColor)
        self.assertEqual(expectedColor, self.getFirstItemTextColor())
        
    def assertBackgroundColor(self):
        self.assertEqual(wx.Colour(*self.newColor), 
                         self.getFirstItemBackgroundColor())
                         
    def assertIcon(self, icon, column=0):
        self.assertEqual(self.viewer.imageIndex[icon], 
                         self.getFirstItemIcon(column))
        

class CommonTestsMixin(object):
    def testCreate(self):
        self.assertItems()

    ## def testCollected(self):
    ##     filterRef = weakref.ref(self.viewer.presentation())
    ##     self.viewer.detach()
    ##     self.viewer = None
    ##     self.parentFrame.Close()
    ##     wx.Yield()
    ##     self.parentFrame = None

    ##     ## import gc
    ##     ## def printRef(obj, indent=0):
    ##     ##     if indent == 3:
    ##     ##         return
    ##     ##     if obj.__class__.__name__ in ['frame']:
    ##     ##         return
    ##     ##     print (' ' * indent), obj.__class__.__name__
    ##     ##     for ref in gc.get_referrers(obj):
    ##     ##         printRef(ref, indent + 1)
    ##     ## print
    ##     ## printRef(filterRef())
    ##     ## print '===', len(gc.get_referrers(filterRef()))

    ##     self.failUnless(filterRef() is None)

    def testAddTask(self):
        self.taskList.append(self.task)
        self.assertItems(self.task)

    def testRemoveTask(self):
        self.taskList.append(self.task)
        self.taskList.remove(self.task)
        self.assertItems()

    def testUndoRemoveTaskWithSubtask(self):
        self.task.addChild(self.child)
        self.taskList.append(self.task)
        self.viewer.select([self.task])
        self.viewer.updateSelection()
        deleteItem = self.viewer.deleteItemCommand()
        deleteItem.do()
        deleteItem.undo()
        if self.viewer.isTreeViewer():
            self.assertItems((self.task, 1), self.child)
        else:
            self.assertItems(self.child, self.task)

    def testDeleteSelectedTask(self):
        self.taskList.append(self.task)
        self.viewer.widget.select_all()
        self.viewer.updateSelection()
        self.taskList.removeItems(self.viewer.curselection())
        self.assertItems()

    def testSelectedTaskStaysSelectedWhenStartingEffortTracking(self):
        self.taskList.append(self.task)
        self.viewer.select([self.task])
        self.assertEqual([self.task], self.viewer.curselection())
        self.task.addEffort(effort.Effort(self.task))
        self.assertEqual([self.task], self.viewer.curselection())
        
    def testChildOrder(self):
        child1 = task.Task(subject='1', plannedStartDateTime=date.Now() - date.ONE_SECOND)
        self.task.addChild(child1)
        child2 = task.Task(subject='2', plannedStartDateTime=date.Now() - date.ONE_SECOND)
        self.task.addChild(child2)
        self.taskList.append(self.task)
        if self.viewer.isTreeViewer():
            self.assertItems((self.task, 2), child1, child2)
        else:
            self.assertItems(child1, child2, self.task)

    def testChildSubjectRendering(self):
        self.task.addChild(self.child)
        self.taskList.append(self.task)
        if self.viewer.isTreeViewer():
            self.assertItems((self.task, 1), self.child)
        else:
            self.assertItems(self.child, self.task)
           
    def testSortOrder(self):
        self.task.addChild(self.child)
        task2 = task.Task(subject='zzz')
        self.taskList.extend([self.task, task2])
        if self.viewer.isTreeViewer():
            self.assertItems((self.task, 1), self.child, task2)
        else:
            self.assertItems(self.child, self.task, task2)

    def testMarkCompleted(self):
        task2 = task.Task(subject='task2')
        self.taskList.extend([self.task, task2])
        self.assertItems(self.task, task2)
        self.task.setCompletionDateTime()
        self.assertItems(task2, self.task)
        
    def testMakeInactive(self):
        task2 = task.Task(subject='task2', plannedStartDateTime=date.Now() - date.ONE_SECOND)
        self.taskList.extend([self.task, task2])
        self.assertItems(self.task, task2)
        self.task.setPlannedStartDateTime(date.Tomorrow())
        self.assertItems(task2, self.task)
        
    def testFilterCompletedTasks(self):
        self.viewer.hideTaskStatus(task.status.completed)
        completedChild = task.Task(completionDateTime=date.Now() - date.ONE_HOUR)
        notCompletedChild = task.Task(plannedStartDateTime=date.Now() - date.ONE_SECOND)
        self.task.addChild(notCompletedChild)
        self.task.addChild(completedChild)
        self.taskList.append(self.task)
        if self.viewer.isTreeViewer():
            self.assertItems((self.task, 1), notCompletedChild)
        else:
            self.assertItems(notCompletedChild, self.task)
            
    def testUndoMarkCompletedWhenFilteringCompletedTasks(self):
        self.settings.setboolean('behavior', 
                                 'markparentcompletedwhenallchildrencompleted', 
                                 True)
        self.viewer.hideTaskStatus(task.status.completed)
        child1 = task.Task('child1')
        child2 = task.Task('child2')
        grandChild = task.Task('grandChild')
        self.task.addChild(child1)
        self.task.addChild(child2)
        child2.addChild(grandChild)
        self.taskList.append(self.task)
        self.viewer.expandAll()
        self.assertEqual(4, self.viewer.size())
        markCompletedCommand = command.MarkCompletedCommand(self.taskList, 
                                                            [grandChild])
        markCompletedCommand.do()
        self.assertEqual(2, self.viewer.size())
        patterns.CommandHistory().undo()
        self.assertEqual(4, self.viewer.size())

    def testFilterOnAllCategories(self):
        self.settings.setboolean('view', 'categoryfiltermatchall', False)
        self.taskList.append(self.task)
        cat1 = category.Category('category 1')
        cat2 = category.Category('category 2')
        self.task.addCategory(cat1)
        cat1.addCategorizable(self.task)
        self.taskFile.categories().extend([cat1, cat2])
        cat1.setFiltered(True)
        cat2.setFiltered(True)
        self.assertEqual(1, self.viewer.size())
        self.settings.setboolean('view', 'categoryfiltermatchall', True)
        self.assertEqual(0, self.viewer.size())
        
    def testFilterOnAnyCategory(self):
        self.settings.setboolean('view', 'categoryfiltermatchall', True)
        self.taskList.append(self.task)
        cat1 = category.Category('category 1')
        cat2 = category.Category('category 2')
        self.task.addCategory(cat1)
        cat1.addCategorizable(self.task)
        self.taskFile.categories().extend([cat1, cat2])
        cat1.setFiltered(True)
        cat2.setFiltered(True)
        self.assertEqual(0, self.viewer.size())
        self.settings.setboolean('view', 'categoryfiltermatchall', False)
        self.assertEqual(1, self.viewer.size())
            
    def testDefaultVisibleColumns(self):
        self.assertEqual(_('Subject'), self.viewer.widget.GetColumn(0).GetText())    
        self.assertEqual(_('Planned start date'), self.viewer.widget.GetColumn(1).GetText())    
        self.assertEqual(_('Due date'), self.viewer.widget.GetColumn(2).GetText())
        self.assertEqual(3, self.viewer.widget.GetColumnCount())
    
    def testTurnOffPlannedStartDateColumn(self):
        self.showColumn('plannedStartDateTime', False)
        self.assertEqual(_('Due date'), self.viewer.widget.GetColumn(1).GetText())
        self.assertEqual(2, self.viewer.widget.GetColumnCount())
        
    def testShowSort_Subject(self):
        self.assertNotEqual(-1, self.viewer.widget.GetColumn(0).GetImage())
        self.assertEqual(-1, self.viewer.widget.GetColumn(1).GetImage())
    
    def testForegroundColorWhenTaskIsCompleted(self):
        self.taskList.append(self.task)
        self.task.setCompletionDateTime()
        newColor = self.task.statusFgColor()
        newColor = wx.Colour(newColor.Red(), newColor.Green(), newColor.Blue())
        self.assertColor(newColor)
        
    def testTurnColumnsOnAndOff(self):
        columns = dict(actualStartDateTime=(3, _('Actual start date')),
                       hourlyFee=(3, _('Hourly fee')), 
                       fixedFee=(3, _('Fixed fee')),
                       revenue=(3, _('Revenue')),
                       priority=(3, _('Priority')),
                       prerequisites=(1, _('Prerequisites')),
                       dependencies=(1, _('Dependents')),
                       categories=(1, _('Categories')),
                       percentageComplete=(3, _('% complete')),
                       recurrence=(3, _('Recurrence')),
                       notes=(1, ''),
                       attachments=(1, ''))
        for column in columns:
            columnIndex, expectedHeader = columns[column] 
            self.showColumn(column)
            actualHeader = self.viewer.widget.GetColumn(columnIndex).GetText()
            self.assertEqual(expectedHeader, actualHeader)
            self.showColumn(column, False)
            self.assertEqual(3, self.viewer.widget.GetColumnCount())
            
    def testRenderFixedFee(self):
        taskWithFixedFee = task.Task(fixedFee=100)
        self.taskList.append(taskWithFixedFee)
        self.showColumn('fixedFee')
        self.assertEqual(locale.currency(100, False), self.getItemText(0, 3))
        self.assertEqual(_('Fixed fee'), 
                         self.viewer.widget.GetColumn(3).GetText())
                        
    def testRenderPercentageComplete_0(self):
        uncompletedTask = task.Task()
        self.taskList.append(uncompletedTask)
        self.showColumn('percentageComplete')
        self.assertEqual('', self.getItemText(0, 3))

    def testRenderPercentageComplete_100(self):
        completedTask = task.Task(completionDateTime=date.Now() - date.ONE_HOUR)
        self.taskList.append(completedTask)
        self.showColumn('percentageComplete')
        self.assertEqual('100%', self.getItemText(0, 3))

    def testRenderSingleCategory(self):
        cat = category.Category(subject='Category')
        self.task.addCategory(cat)
        cat.addCategorizable(self.task)
        self.assertEqual('Category', self.viewer.renderCategories(self.task))
                
    def testRenderMultipleCategories(self):
        for index in range(1, 3):
            cat = category.Category(subject='Category %d' % index)
            self.task.addCategory(cat)
            cat.addCategorizable(self.task)
        self.assertEqual('Category 1, Category 2', self.viewer.renderCategories(self.task))
        
    def testRenderSingleChildCategory(self):
        self.task.addChild(self.child)
        self.taskList.append(self.task)
        cat = category.Category(subject='Category')
        self.child.addCategory(cat)
        cat.addCategorizable(self.child)
        expectedCategory = '(Category)' if self.viewer.isTreeViewer() else ''
        self.assertEqual(expectedCategory, self.viewer.renderCategories(self.task))

    def testRenderMultipleChildCategories(self):
        self.task.addChild(self.child)
        self.taskList.append(self.task)
        for index in range(1, 3):
            cat = category.Category(subject='Category %d' % index)
            self.child.addCategory(cat)
            cat.addCategorizable(self.child)
        expectedCategory = '(Category 1, Category 2)' if self.viewer.isTreeViewer() else ''
        self.assertEqual(expectedCategory, self.viewer.renderCategories(self.task))
        
    def testRenderDifferentParentAndChildCategories(self):
        self.task.addChild(self.child)
        self.taskList.append(self.task)
        for index, eachTask in enumerate([self.task, self.child]):
            cat = category.Category(subject='Category %d' % index)
            eachTask.addCategory(cat)
            cat.addCategorizable(eachTask)
        expectedCategory = 'Category 0 (Category 1)' if self.viewer.isTreeViewer() else 'Category 0'
        self.assertEqual(expectedCategory, self.viewer.renderCategories(self.task))

    def testRenderSameParentAndChildCategory(self):
        self.task.addChild(self.child)
        self.taskList.append(self.task)
        cat = category.Category(subject='Category')
        for eachTask in (self.task, self.child):
            eachTask.addCategory(cat)
            cat.addCategorizable(eachTask)
        expectedCategory = 'Category'
        self.assertEqual(expectedCategory, self.viewer.renderCategories(self.task))

    def testRenderRecurrence(self):
        taskWithRecurrence = task.Task(recurrence=date.Recurrence('weekly', amount=2))
        self.showColumn('recurrence')
        self.taskList.append(taskWithRecurrence)
        self.assertEqual('Every other week', self.getItemText(0, 3))
        
    def testRenderAttachment(self):
        att = attachment.FileAttachment('whatever')
        self.task.addAttachment(att)
        self.taskList.append(self.task)
        self.showColumn('attachments')
        self.assertIcon('paperclip_icon', column=1)
        
    def testOneDayLeft(self):
        self.showColumn('timeLeft')
        timeLeft = date.TimeDelta(hours=25, seconds=30)
        self.taskList.append(self.task)
        self.task.setDueDateTime(date.Now() + timeLeft)
        self.assertEqual(render.timeLeft(timeLeft, False), 
                         self.getItemText(0, 3))
        
    def testReverseSortOrderWithGrandchildren(self):
        self.task.addChild(self.child)
        grandchild = task.Task(subject='grandchild', plannedStartDateTime=date.Now() - date.ONE_SECOND)
        self.child.addChild(grandchild)
        task2 = task.Task(subject='zzz', plannedStartDateTime=date.Now() - date.ONE_SECOND)
        self.taskList.extend([self.task, task2])
        self.viewer.setSortOrderAscending(False)
        if self.viewer.isTreeViewer():
            self.assertItems(task2, (self.task, 1), (self.child, 1), grandchild)
        else:
            self.assertItems(task2, self.task, grandchild, self.child)
                
    def testReverseSortOrder(self):
        self.task.addChild(self.child)
        task2 = task.Task(subject='zzz', plannedStartDateTime=date.Now() - date.ONE_SECOND)
        self.taskList.extend([self.task, task2])
        self.viewer.setSortOrderAscending(False)
        if self.viewer.isTreeViewer():
            self.assertItems(task2, (self.task, 1), self.child)
        else:
            self.assertItems(task2, self.task, self.child)

    def testSortByDueDate(self):
        self.task.addChild(self.child)
        task2 = task.Task(subject='zzz', plannedStartDateTime=date.Now() - date.ONE_SECOND)
        child2 = task.Task(subject='child 2', plannedStartDateTime=date.Now() - date.ONE_SECOND)
        task2.addChild(child2)
        child2.setParent(task2)
        self.taskList.extend([self.task, task2])
        if self.viewer.isTreeViewer():
            self.assertItems((self.task, 1), self.child, (task2, 1), child2)
        else:
            self.assertItems(self.child, child2, self.task, task2) 
        child2.setDueDateTime(date.Now().endOfDay())
        self.viewer.sortBy('dueDateTime')
        if self.viewer.isTreeViewer():
            self.assertItems((task2, 1), child2, (self.task, 1), self.child)
        else:    
            self.assertItems(child2, self.child, self.task, task2)
            
    def testSortByPrerequisite_OnePrerequisite(self):
        self.viewer.sortBy('prerequisites')
        prerequisite = task.Task()
        self.task.addPrerequisites([prerequisite])
        self.taskList.extend([prerequisite, self.task])
        self.assertItems(prerequisite, self.task)

    def testSortByPrerequisite_TwoPrerequisites(self):
        self.viewer.sortBy('prerequisites')
        prerequisite1 = task.Task(subject='1')
        prerequisite2 = task.Task(subject='2')
        self.task.addPrerequisites([prerequisite1, prerequisite2])
        self.taskList.extend([prerequisite1, prerequisite2, self.task])
        try:
            self.assertItems(prerequisite1, prerequisite2, self.task)
        except AssertionError:  # pragma: no cover
            self.assertItems(prerequisite2, prerequisite1, self.task)

    def testSortByPrerequisite_ChainedPrerequisites(self):
        self.viewer.sortBy('prerequisites')
        task0 = task.Task(subject='0')
        task1 = task.Task(subject='1')
        task2 = task.Task(subject='2')
        task2.addPrerequisites([task1])
        task1.addPrerequisites([task0])
        self.taskList.extend([task0, task1, task2])
        self.assertItems(task0, task1, task2)  # Prerequisites = '', '0', '1'
        self.viewer.setSortOrderAscending(False)
        self.assertItems(task2, task1, task0)  # Prerequisites = '1', '0', ''

    def testSortBySubject_AddPrerequisite(self):
        task0 = task.Task(subject='0', plannedStartDateTime=date.DateTime(2000, 1, 1))
        task1 = task.Task(subject='1', plannedStartDateTime=date.DateTime(2000, 1, 1))
        self.taskList.extend([task0, task1])
        self.assertItems(task0, task1)
        task0.addPrerequisites([task1])
        self.assertItems(task1, task0)
        
    def testSortByCategories(self):
        cat0 = category.Category(subject='Category 0')
        cat1 = category.Category(subject='Category 1')
        task0 = task.Task(subject='0')
        task1 = task.Task(subject='1')
        task0.addCategory(cat1)
        cat1.addCategorizable(task0)
        task1.addCategory(cat0)
        cat0.addCategorizable(task1)
        self.taskList.extend([task0, task1])
        self.assertItems(task0, task1)
        self.viewer.sortBy('categories')
        self.assertItems(task1, task0)

    def testSortByChildCategories(self):
        cat0 = category.Category(subject='Category 0')
        cat1 = category.Category(subject='Category 1')
        task0 = task.Task(subject='0')
        task1 = task.Task(subject='1')
        task1_1 = task.Task(subject='1.1')
        task1.addChild(task1_1)
        task0.addCategory(cat1)
        cat1.addCategorizable(task0)
        task1_1.addCategory(cat0)
        cat0.addCategorizable(task1_1)
        self.taskList.extend([task0, task1])
        if self.viewer.isTreeViewer():
            self.assertItems(task0, (task1, 1), task1_1)
        else:
            self.assertItems(task0, task1, task1_1)
        self.viewer.sortBy('categories')
        if self.viewer.isTreeViewer():
            self.assertItems((task1, 1), task1_1, task0)
        else:
            self.assertItems(task1, task1_1, task0)
        
    def testChangeActiveTaskForegroundColor(self):
        self.setColor('activetasks')
        self.taskList.append(task.Task(subject='test', actualStartDateTime=date.Now()))
        self.assertColor()
    
    def testChangeInactiveTaskForegroundColor(self):
        self.setColor('inactivetasks')
        self.taskList.append(task.Task())
        self.assertColor()
    
    def testChangeCompletedTaskForegroundColor(self):
        self.setColor('completedtasks')
        self.taskList.append(task.Task(completionDateTime=date.Now()))
        self.assertColor()

    def testChangeDueSoonTaskForegroundColor(self):
        self.setColor('duesoontasks')
        self.taskList.append(task.Task(dueDateTime=date.Now().endOfDay()))
        self.assertColor()

    def testChangeOverDueTaskForegroundColor(self):
        self.setColor('overduetasks')
        self.taskList.append(task.Task(dueDateTime=date.Yesterday()))
        self.assertColor()
            
    def testStatusMessage_EmptyTaskList(self):
        self.assertEqual(('Tasks: 0 selected, 0 visible, 0 total', 
            'Status: 0 overdue, 0 late, 0 inactive, 0 completed'),
            self.viewer.statusMessages())
    
    def testOnDropFiles(self):
        aTask = task.Task()
        self.taskList.append(aTask)
        self.viewer.onDropFiles(aTask, ['filename'])
        self.assertEqual([attachment.FileAttachment('filename')],
                         self.viewer.presentation()[0].attachments())

    def testOnDropURL(self):
        aTask = task.Task()
        self.taskList.append(aTask)
        self.viewer.onDropURL(aTask, 'http://www.example.com/')
        self.assertEqual([attachment.URIAttachment('http://www.example.com/')],
                         self.viewer.presentation()[0].attachments())

    def testOnDropMail(self):
        file('test.mail', 'wb').write('Subject: foo\r\n\r\nBody\r\n')
        aTask = task.Task()
        self.taskList.append(aTask)
        self.viewer.onDropMail(aTask, 'test.mail')
        self.assertEqual([attachment.MailAttachment('test.mail')],
                         self.viewer.presentation()[0].attachments())
        
    def testCategoryBackgroundColor(self):
        cat = category.Category('category with background color', bgColor=self.newColor)
        cat.addCategorizable(self.task)
        self.task.addCategory(cat)
        self.taskList.append(self.task)
        self.assertBackgroundColor()
        
    def testNewItem(self):
        self.taskFile.categories().append(category.Category('cat', filtered=True))
        dialog = self.viewer.newItemDialog(bitmap='new')
        dialog._interior[4].selected()
        tree = dialog._interior[4].viewer.widget  # pylint: disable=W0212
        firstChild = tree.GetFirstChild(tree.GetRootItem())[0]
        self.failUnless(firstChild.IsChecked())
        
    def testFont(self):
        self.taskList.append(task.Task(font=wx.SWISS_FONT))
        self.assertEqual(wx.SWISS_FONT, self.getFirstItemFont())
        
    def testIconUpdatesWhenPlannedStartDateTimeChanges(self):
        self.taskList.append(self.task)
        self.task.setPlannedStartDateTime(date.Now() + date.ONE_DAY)
        self.assertIcon(task.inactive.getBitmap(self.settings))

    def testIconUpdatesWhenDueDateTimeChanges(self):
        self.taskList.append(self.task)
        self.task.setDueDateTime(date.Now() + date.ONE_HOUR)
        self.assertIcon(task.duesoon.getBitmap(self.settings))

    def testIconUpdatesWhenCompletionDateTimeChanges(self):
        self.taskList.append(self.task)
        self.task.setCompletionDateTime(date.Now())
        self.assertIcon(task.completed.getBitmap(self.settings))

    def testIconUpdatesWhenPrerequisiteIsAdded(self):
        prerequisite = task.Task('zzz')
        self.taskList.extend([prerequisite, self.task])
        self.task.addPrerequisites([prerequisite])
        prerequisite.addDependencies([self.task])
        self.assertIcon(task.inactive.getBitmap(self.settings))

    def testIconUpdatesWhenPrerequisiteIsCompleted(self):
        prerequisite = task.Task(subject='zzz')
        self.taskList.extend([prerequisite, self.task])
        self.task.addPrerequisites([prerequisite])
        prerequisite.addDependencies([self.task]) 
        prerequisite.setCompletionDateTime(date.Now())
        self.assertIcon(task.late.getBitmap(self.settings))
        
    def testIconUpdatesWhenEffortTrackingStarts(self):
        self.taskList.append(self.task)
        self.task.addEffort(effort.Effort(self.task))
        self.assertIcon('clock_icon')
        
    def testIconUpdatesWhenEffortTrackingStops(self):
        self.taskList.append(self.task)
        self.task.addEffort(effort.Effort(self.task))
        self.task.stopTracking()
        self.assertIcon(task.active.getBitmap(self.settings))
        
    def testIconUpdatesWhenTaskBecomesOverdue(self):
        dueDateTime = date.Now() + date.TimeDelta(seconds=10)
        dueDateTime = dueDateTime.replace(microsecond=0)
        self.task.setDueDateTime(dueDateTime)
        self.taskList.append(self.task)
        self.assertIcon(task.duesoon.getBitmap(self.settings))
        now = dueDateTime + date.ONE_SECOND
        oldNow = date.Now
        date.Now = lambda: now
        self.task.onOverDue()        
        self.assertIcon(task.overdue.getBitmap(self.settings))
        date.Now = oldNow
        
    def testModeIsSavedInSettings(self):
        self.assertEqual(self.treeMode, 
            self.settings.getboolean(self.viewer.settingsSection(), 'treemode'))

    def testRenderSubject(self):
        self.task.addChild(self.child)
        expectedSubject = 'child' if self.treeMode else 'task -> child'
        self.assertEqual(expectedSubject, self.viewer.renderSubject(self.child))

    def testItemOrder(self):
        self.task.addChild(self.child)
        self.taskList.append(self.task)        
        if self.treeMode:
            self.assertItems((self.task, 1), self.child)
        else:
            self.assertItems(self.child, self.task)
                         
    def testItemOrderAfterSwitch(self):
        self.task.addChild(self.child)
        self.taskList.append(self.task)
        self.settings.setboolean(self.viewer.settingsSection(), 'treemode',
                                 not self.treeMode)
        if self.treeMode:
            self.assertItems(self.child, self.task)
        else:
            self.assertItems((self.task, 1), self.child)

    def testItemOrderAfterSwitchWhenOrderDoesNotChange(self):
        self.task.addChild(self.child)
        self.taskList.append(self.task)
        self.task.setSubject('a')  # task comes before child
        self.settings.setboolean(self.viewer.settingsSection(), 'treemode',
                                 not self.treeMode)
        if self.treeMode:
            self.assertItems(self.task, self.child)
        else:
            self.assertItems((self.task, 1), self.child)
            
    def assertEventFired_Deprecated(self, type_):
        types = []
        for event in self.viewer.events_deprecated:
            types.extend(event.types())
        self.failUnless(type_ in types,
                        '"%s" not in %s' % (type_, self.viewer.events_deprecated))
        
    def assertEventFired(self, newValue, sender):
        self.failUnless((newValue, sender) in self.viewer.events)

    def testGetTimeSpent(self):
        self.taskList.append(self.task)
        self.task.addEffort(effort.Effort(self.task, date.DateTime(2000, 1, 1),
                                                     date.DateTime(2000, 1, 2)))
        self.showColumn('timeSpent')
        timeSpent = self.getItemText(0, 3)
        self.assertEqual("24:00:00", timeSpent)

    def testGetTotalTimeSpent(self):
        self.task.addChild(self.child)
        self.taskList.append(self.task)
        self.task.expand(False, context=self.viewer.settingsSection())
        self.task.addEffort(effort.Effort(self.task, date.DateTime(2000, 1, 1),
                                                     date.DateTime(2000, 1, 2)))
        self.child.addEffort(effort.Effort(self.child, date.DateTime(2000, 1, 1),
                                                     date.DateTime(2000, 1, 2)))
        self.showColumn('timeSpent')
        timeSpent = self.getItemText(0, 3)
        expectedTimeSpent = "(48:00:00)" if self.treeMode else "24:00:00"
        self.assertEqual(expectedTimeSpent, timeSpent)
        
    def testGetSelection(self):
        taskA = task.Task('a')
        taskB = task.Task('b')
        self.viewer.presentation().extend([taskA, taskB])
        self.viewer.select([taskA])
        self.viewer.updateSelection()
        self.assertEqual([taskA], self.viewer.curselection())

    def testGetSelection_AfterResort(self):
        taskA = task.Task('a')
        taskB = task.Task('b')
        self.viewer.presentation().extend([taskA, taskB])
        self.viewer.widget.select([taskA])
        self.viewer.updateSelection()
        self.viewer.setSortOrderAscending(False)
        self.assertEqual([taskA], self.viewer.curselection())
        
    def testChangeSubject(self):
        self.taskList.append(self.task)
        self.task.setSubject('New subject')
        self.assertEqual(task.Task.subjectChangedEventType(), 
                         self.viewer.events_deprecated[0].type())

    def testChangePlannedStartDateTimeWhileColumnShown(self):
        self.taskList.append(self.task)
        newValue = date.Now() - date.ONE_DAY
        self.task.setPlannedStartDateTime(newValue)
        self.assertEqual((newValue, self.task), self.viewer.events[0])

    def testStartTracking(self):
        self.taskList.append(self.task)
        self.task.addEffort(effort.Effort(self.task))
        self.failUnless((True, self.task) in self.viewer.events)

    def testChangePlannedStartDateTimeWhileColumnNotShown(self):
        self.taskList.append(self.task)
        self.showColumn('plannedStartDate', False)
        self.task.setPlannedStartDateTime(date.Yesterday())
        self.assertEqual(1, len(self.viewer.events))

    def testChangeDueDate(self):
        self.taskList.append(self.task)
        newValue = date.Now().endOfDay()
        self.task.setDueDateTime(newValue)
        self.failUnless((newValue, self.task) in self.viewer.events)

    def testChangeCompletionDateWhileColumnNotShown(self):
        self.taskList.append(self.task)
        now = date.Now()
        self.task.setCompletionDateTime(now)
        # We still get an event for the subject column:
        self.failUnless((now, self.task) in self.viewer.events)

    def testChangeCompletionDateWhileColumnShown(self):
        self.taskList.append(self.task)
        self.showColumn('completionDate')
        now = date.Now()
        self.task.setCompletionDateTime(now)
        self.failUnless((now, self.task) in self.viewer.events)

    def testChangePercentageCompleteWhileColumnNotShown(self):
        self.taskList.append(self.task)
        self.task.setPercentageComplete(50)
        self.failIf((50, self.task) in self.viewer.events)

    def testChangePercentageCompleteWhileColumnShown(self):
        self.taskList.append(self.task)
        self.showColumn('percentageComplete')
        self.task.setPercentageComplete(50)
        self.failUnless((50, self.task) in self.viewer.events)

    def testChangePriorityWhileColumnNotShown(self):
        self.taskList.append(self.task)
        self.task.setPriority(10)
        self.failIf(self.viewer.events)

    def testChangePriorityWhileColumnShown(self):
        self.taskList.append(self.task)
        self.showColumn('priority')
        self.task.setPriority(10)
        self.assertEventFired(10, self.task)

    def testChangePriorityOfSubtask(self):
        self.showColumn('priority')
        self.task.addChild(self.child)
        self.taskList.append(self.task)
        self.child.setPriority(10)
        self.assertEventFired(self.task.priority(), self.task)
        
    def testChangeHourlyFeeWhileColumnShown(self):
        self.showColumn('hourlyFee')
        self.taskList.append(self.task)
        self.task.setHourlyFee(100)
        self.assertEqual(render.monetaryAmount(100.), self.getItemText(0, 3))
        
    def testChangeFixedFeeWhileColumnShown(self):
        self.showColumn('fixedFee')
        self.taskList.append(self.task)
        self.task.setFixedFee(200)
        self.assertEqual(render.monetaryAmount(200.), self.getItemText(0, 3))

    def testCollapsedCompositeTaskShowsRecursiveFixedFee(self):
        self.showColumn('fixedFee')
        self.taskList.extend([self.task, self.child])
        self.task.addChild(self.child)
        self.task.setFixedFee(100)
        self.child.setFixedFee(200)
        self.viewer.setSortOrderAscending(False)
        expectedAmount = "(%s)" % locale.currency(300, False) if self.treeMode else locale.currency(100, False)
        self.task.expand(False, context=self.viewer.settingsSection())
        self.assertEqual(expectedAmount, self.getItemText(0, 3))
        
    def testCollapsedCompositeTaskShowsRecursivePlannedStartDateTime(self):
        self.taskList.extend([self.task, self.child])
        self.task.addChild(self.child)
        now = date.Now()
        self.child.setPlannedStartDateTime(now)
        self.task.setPlannedStartDateTime(date.DateTime())
        self.viewer.setSortByTaskStatusFirst(False)
        self.viewer.setSortOrderAscending(False)
        expectedDateTime = "(%s)" % render.dateTime(now, humanReadable=True) if self.treeMode else ''
        self.task.expand(False, context=self.viewer.settingsSection())
        self.assertEqual(expectedDateTime, self.getItemText(0, 1))

    def testChangePrerequisiteSubject(self):
        self.showColumn('prerequisites')
        self.viewer.setSortOrderAscending(False)
        prerequisite = task.Task(subject='prerequisite')
        self.taskList.extend([self.task, prerequisite])
        self.task.addPrerequisites([prerequisite])
        prerequisite.addDependencies([self.task])
        self.assertEqual('prerequisite', self.getItemText(0, 1))
        prerequisite.setSubject('new')
        self.assertEqual('new', self.getItemText(0, 1))

    def testChangeDependencySubject(self):
        self.showColumn('dependencies')
        self.viewer.setSortOrderAscending(False)
        dependency = task.Task(subject='dependency')
        self.taskList.extend([self.task, dependency])
        dependency.addPrerequisites([self.task])
        self.task.addDependencies([dependency])
        self.assertEqual('dependency', self.getItemText(0, 1))
        dependency.setSubject('new')
        self.assertEqual('new', self.getItemText(0, 1))

    def testPlannedStartDateTimeToday(self):
        today = date.Now()
        self.task.setPlannedStartDateTime(today)
        self.taskList.append(self.task)
        self.showColumn('plannedStartDateTime')
        self.assertEqual(_('Today %s') % render.time(today.time()), self.getItemText(0, 1))

    def testPlannedStartDateTimeYesterday(self):
        yesterday = date.Yesterday()
        self.task.setPlannedStartDateTime(yesterday)
        self.taskList.append(self.task)
        self.showColumn('plannedStartDateTime')
        self.assertEqual(_('Yesterday %s') % render.time(yesterday.time()), self.getItemText(0, 1))

    def testPlannedStartDateTimeTomorrow(self):
        tomorrow = date.Tomorrow()
        self.task.setPlannedStartDateTime(tomorrow)
        self.taskList.append(self.task)
        self.showColumn('plannedStartDateTime')
        self.assertEqual(_('Tomorrow %s') % render.time(tomorrow.time()), self.getItemText(0, 1))

    def testPlannedStartDateToday(self):
        today = date.Now().startOfDay()
        self.task.setPlannedStartDateTime(today)
        self.taskList.append(self.task)
        self.showColumn('plannedStartDateTime')
        self.assertEqual(_('Today'), self.getItemText(0, 1))

    def testPlannedStartDateYesterday(self):
        yesterday = date.Yesterday().startOfDay()
        self.task.setPlannedStartDateTime(yesterday)
        self.taskList.append(self.task)
        self.showColumn('plannedStartDateTime')
        self.assertEqual(_('Yesterday'), self.getItemText(0, 1))

    def testPlannedStartDateTomorrow(self):
        tomorrow = date.Tomorrow().startOfDay()
        self.task.setPlannedStartDateTime(tomorrow)
        self.taskList.append(self.task)
        self.showColumn('plannedStartDateTime')
        self.assertEqual(_('Tomorrow'), self.getItemText(0, 1))

    def testDueDateTimeToday(self):
        today = date.Now()
        self.task.setDueDateTime(today)
        self.taskList.append(self.task)
        self.showColumn('dueDateTime')
        self.assertEqual(_('Today %s') % render.time(today.time()), self.getItemText(0, 2))

    def testDueDateTimeYesterday(self):
        yesterday = date.Yesterday()
        self.task.setDueDateTime(yesterday)
        self.taskList.append(self.task)
        self.showColumn('dueDateTime')
        self.assertEqual(_('Yesterday %s') % render.time(yesterday.time()), self.getItemText(0, 2))

    def testDueDateTimeTomorrow(self):
        tomorrow = date.Tomorrow()
        self.task.setDueDateTime(tomorrow)
        self.taskList.append(self.task)
        self.showColumn('dueDateTime')
        self.assertEqual(_('Tomorrow %s') % render.time(tomorrow.time()), self.getItemText(0, 2))

    def testDueDateToday(self):
        today = date.Now().startOfDay()
        self.task.setDueDateTime(today)
        self.taskList.append(self.task)
        self.showColumn('dueDateTime')
        self.assertEqual(_('Today'), self.getItemText(0, 2))

    def testDueDateYesterday(self):
        yesterday = date.Yesterday().startOfDay()
        self.task.setDueDateTime(yesterday)
        self.taskList.append(self.task)
        self.showColumn('dueDateTime')
        self.assertEqual(_('Yesterday'), self.getItemText(0, 2))

    def testDueDateTomorrow(self):
        tomorrow = date.Tomorrow().startOfDay()
        self.task.setDueDateTime(tomorrow)
        self.taskList.append(self.task)
        self.showColumn('dueDateTime')
        self.assertEqual(_('Tomorrow'), self.getItemText(0, 2))

    def testActualStartDateTimeToday(self):
        today = date.Now()
        self.task.setActualStartDateTime(today)
        self.taskList.append(self.task)
        self.showColumn('actualStartDateTime')
        self.assertEqual(_('Today %s') % render.time(today.time()), self.getItemText(0, 3))

    def testActualStartDateTimeYesterday(self):
        yesterday = date.Yesterday()
        self.task.setActualStartDateTime(yesterday)
        self.taskList.append(self.task)
        self.showColumn('actualStartDateTime')
        self.assertEqual(_('Yesterday %s') % render.time(yesterday.time()), self.getItemText(0, 3))

    def testActualStartDateTimeTomorrow(self):
        tomorrow = date.Tomorrow()
        self.task.setActualStartDateTime(tomorrow)
        self.taskList.append(self.task)
        self.showColumn('actualStartDateTime')
        self.assertEqual(_('Tomorrow %s') % render.time(tomorrow.time()), self.getItemText(0, 3))

    def testActualStartDateToday(self):
        today = date.Now().startOfDay()
        self.task.setActualStartDateTime(today)
        self.taskList.append(self.task)
        self.showColumn('actualStartDateTime')
        self.assertEqual(_('Today'), self.getItemText(0, 3))

    def testActualStartDateYesterday(self):
        yesterday = date.Yesterday().startOfDay()
        self.task.setActualStartDateTime(yesterday)
        self.taskList.append(self.task)
        self.showColumn('actualStartDateTime')
        self.assertEqual(_('Yesterday'), self.getItemText(0, 3))

    def testActualStartDateTomorrow(self):
        tomorrow = date.Tomorrow().startOfDay()
        self.task.setActualStartDateTime(tomorrow)
        self.taskList.append(self.task)
        self.showColumn('actualStartDateTime')
        self.assertEqual(_('Tomorrow'), self.getItemText(0, 3))

    def testCompletionDateTimeToday(self):
        today = date.Now()
        self.task.setCompletionDateTime(today)
        self.taskList.append(self.task)
        self.showColumn('completionDateTime')
        self.assertEqual(_('Today %s') % render.time(today.time()), self.getItemText(0, 3))

    def testCompletionDateTimeYesterday(self):
        yesterday = date.Yesterday()
        self.task.setCompletionDateTime(yesterday)
        self.taskList.append(self.task)
        self.showColumn('completionDateTime')
        self.assertEqual(_('Yesterday %s') % render.time(yesterday.time()), self.getItemText(0, 3))

    def testCompletionDateTimeTomorrow(self):
        tomorrow = date.Tomorrow()
        self.task.setCompletionDateTime(tomorrow)
        self.taskList.append(self.task)
        self.showColumn('completionDateTime')
        self.assertEqual(_('Tomorrow %s') % render.time(tomorrow.time()), self.getItemText(0, 3))

    def testCompletionDateToday(self):
        today = date.Now().startOfDay()
        self.task.setCompletionDateTime(today)
        self.taskList.append(self.task)
        self.showColumn('completionDateTime')
        self.assertEqual(_('Today'), self.getItemText(0, 3))

    def testCompletionDateYesterday(self):
        yesterday = date.Yesterday().startOfDay()
        self.task.setCompletionDateTime(yesterday)
        self.taskList.append(self.task)
        self.showColumn('completionDateTime')
        self.assertEqual(_('Yesterday'), self.getItemText(0, 3))

    def testCompletionDateTomorrow(self):
        tomorrow = date.Tomorrow().startOfDay()
        self.task.setCompletionDateTime(tomorrow)
        self.taskList.append(self.task)
        self.showColumn('completionDateTime')
        self.assertEqual(_('Tomorrow'), self.getItemText(0, 3))

    # Test all attributes...


class TaskViewerInTreeModeTest(CommonTestsMixin, TaskViewerTestCase):
    treeMode = True


class TaskViewerInListModeTest(CommonTestsMixin, TaskViewerTestCase):
    treeMode = False
        

class TaskCalendarViewerTest(test.wxTestCase):
    def setUp(self):
        super(TaskCalendarViewerTest, self).setUp()
        task.Task.settings = self.settings = config.Settings(load=False)
        self.taskFile = persistence.TaskFile()
        self.frame.taskFile = self.taskFile
        self.viewer = gui.viewer.task.CalendarViewer(self.frame, self.taskFile, 
                                                     self.settings)
        self.originalTopWindow = wx.GetApp().TopWindow
        wx.GetApp().TopWindow = self.frame  # uiCommands use TopWindow to get the main window
        
    def tearDown(self):
        super(TaskCalendarViewerTest, self).tearDown()
        wx.GetApp().TopWindow = self.originalTopWindow
        self.taskFile.close()
        self.taskFile.stop()
        
    def openDialogAndAssertDateTimes(self, dateTime, expectedPlannedStartDateTime, 
                                     expectedDueDateTime):
        self.viewer.onCreate(dateTime, show=False)
        newTask = list(self.taskFile.tasks())[0]
        self.assertEqual(expectedPlannedStartDateTime, newTask.plannedStartDateTime())
        self.assertEqual(expectedDueDateTime, newTask.dueDateTime())
        
    def testOnCreateSetsPlannedStartAndDueDateTime(self):
        dateTime = date.DateTime(2010, 10, 10, 16, 0, 0)
        self.openDialogAndAssertDateTimes(dateTime, dateTime, dateTime)

    def testOnCreateKeepsPlannedStartDateTimeAndMakesDueDateTimeEndOfDayWhenDateTimeIsStartOfDay(self):
        dateTime = date.DateTime(2010, 10, 1, 0, 0, 0)
        self.openDialogAndAssertDateTimes(dateTime, dateTime, dateTime.endOfDay())

        
class TaskSquareMapViewerTest(test.wxTestCase):
    def testCreate(self):
        task.Task.settings = settings = config.Settings(load=False)
        self.taskFile = persistence.TaskFile()
        gui.viewer.task.SquareTaskViewer(self.frame, self.taskFile, settings)

    def tearDown(self):
        super(TaskSquareMapViewerTest, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()


class TaskTimelineViewerTest(test.wxTestCase):
    def testCreate(self):
        # pylint: disable-msg=W0201
        task.Task.settings = settings = config.Settings(load=False)
        self.taskFile = persistence.TaskFile()
        gui.viewer.task.TimelineViewer(self.frame, self.taskFile, settings)

    def tearDown(self):
        super(TaskTimelineViewerTest, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()


class TaskStatisticsViewerTest(test.wxTestCase):
    def testCreate(self):
        task.Task.settings = settings = config.Settings(load=False)
        self.taskFile = persistence.TaskFile()
        gui.viewer.task.TaskStatsViewer(self.frame, self.taskFile, settings)

    def tearDown(self):
        super(TaskStatisticsViewerTest, self).tearDown()
        self.taskFile.close()
        self.taskFile.stop()



#!/usr/bin/env python
# -*- coding: utf-8 -*-

from wxSchedulerConstants import *
from wxScheduleUtils import copyDateTime
from wxTimeFormat import wxTimeFormat

import wx, math


class wxDrawer(object):
	"""
	This class handles the actual painting of headers and schedules.
	"""

	# Set this to True if you want your methods to be passed a
	# wx.GraphicsContext instead of wx.DC.
	use_gc = False

	def __init__(self, context, displayedHours):
		self.context = context
		self.displayedHours = displayedHours

	def AdjustFontForHeight(self, font, height):
		pointSize = 18
		while True:
			font.SetPointSize( pointSize )
			_, th = self.context.GetTextExtent(' ' + wxTimeFormat.FormatTime( wx.DateTimeFromHMS(23, 59, 59) ))
			if th <= height:
				return
			pointSize -= 1
			if pointSize == 1:
				return # Hum

	def AdjustFontForWidth(self, font, width):
		pointSize = 18
		while True:
			font.SetPointSize( pointSize )
			self.context.SetFont( font )
			tw, _ = self.context.GetTextExtent(' ' + wxTimeFormat.FormatTime( wx.DateTimeFromHMS(23, 59, 59) ))
			if tw <= width:
				return
			pointSize -= 1
			if pointSize == 1:
				return # Hum

	def DrawDayHeader(self, day, x, y, w, h, highlight=None):
		"""
		Draws the header for a day. Returns the header's size.
		"""
		raise NotImplementedError

	def DrawDayBackground(self, x, y, w, h, highlight=None):
		"""
		Draws the background for a day.
		"""
		raise NotImplementedError

	def DrawMonthHeader(self, day, x, y, w, h):
		"""
		Draws the header for a month. Returns the header's size.
		"""
		raise NotImplementedError

	def DrawSimpleDayHeader(self, day, x, y, w, h, highlight=None):
		"""
		Draws the header for a day, in compact form. Returns
		the header's size.
		"""
		raise NotImplementedError

	def DrawHours(self, x, y, w, h, direction, includeText=True):
		"""
		Draws hours of the day on the left of the specified
		rectangle. Returns the days column size.
		"""
		raise NotImplementedError

	def DrawSchedulesCompact(self, day, schedules, x, y, width, height, highlightColor):
		"""
		Draws a set of schedules in compact form (vertical
		month). Returns a list of (schedule, point, point).
		"""
		raise NotImplementedError

	def DrawNowHorizontal(self, x, y, w):
		"""
		Draws a horizontal line showing when is now
		"""
		raise NotImplementedError

	def DrawNowVertical(self, x, y, h):
		"""
		Draws a vertical line showing when is now
		"""
		raise NotImplementedError

	def _DrawSchedule(self, schedule, x, y, w, h):
		"""
		Draws a schedule in the specified rectangle.
		"""

		offsetY = SCHEDULE_INSIDE_MARGIN
		offsetX = SCHEDULE_INSIDE_MARGIN

		if self.use_gc:
			if h is not None:
				pen = wx.Pen(schedule.color)
				self.context.SetPen(self.context.CreatePen(pen))

				brush = self.context.CreateLinearGradientBrush(x, y, x + w, y + h, schedule.color, SCHEDULER_BACKGROUND_BRUSH())
				self.context.SetBrush(brush)
				self.context.DrawRoundedRectangle(x, y, w, h, SCHEDULE_INSIDE_MARGIN)

			if schedule.complete is not None:
				if h is not None:
					self.context.SetPen(self.context.CreatePen(wx.Pen(wx.SystemSettings.GetColour(wx.SYS_COLOUR_SCROLLBAR))))
					self.context.SetBrush(self.context.CreateBrush(wx.Brush(wx.SystemSettings.GetColour(wx.SYS_COLOUR_SCROLLBAR))))
					self.context.DrawRoundedRectangle(x + SCHEDULE_INSIDE_MARGIN, y + offsetY,
									  w - 2 * SCHEDULE_INSIDE_MARGIN, 2 * SCHEDULE_INSIDE_MARGIN, SCHEDULE_INSIDE_MARGIN)

					if schedule.complete:
						self.context.SetBrush(self.context.CreateLinearGradientBrush(x + SCHEDULE_INSIDE_MARGIN, y + offsetY,
													     x + (w - 2 * SCHEDULE_INSIDE_MARGIN) * schedule.complete,
													     y + offsetY + 10,
													     wx.Colour(0, 0, 255),
													     wx.Colour(0, 255, 255)))
						self.context.DrawRoundedRectangle(x + SCHEDULE_INSIDE_MARGIN, y + offsetY,
										  (w - 2 * SCHEDULE_INSIDE_MARGIN) * schedule.complete, 10, 5)

				offsetY += 10 + SCHEDULE_INSIDE_MARGIN

			if schedule.icons:
				for icon in schedule.icons:
					if h is not None:
						bitmap = wx.ArtProvider.GetBitmap( icon, wx.ART_FRAME_ICON, (16, 16) )
						self.context.DrawBitmap( bitmap, x + offsetX, y + offsetY, 16, 16 )
					offsetX += 20
					if offsetX > w - SCHEDULE_INSIDE_MARGIN:
						offsetY += 20
						offsetX = SCHEDULE_INSIDE_MARGIN
						break

			font = schedule.font
			self.context.SetFont(font, schedule.foreground)
			offsetY += self._drawTextInRect( self.context, schedule.description, offsetX,
							 x, y + offsetY, w - 2 * SCHEDULE_INSIDE_MARGIN, None if h is None else h - offsetY - SCHEDULE_INSIDE_MARGIN )
		else:
			if h is not None:
				self.context.SetBrush(wx.Brush(schedule.color))
				self.context.DrawRectangle(x, y, w, h)

			if schedule.complete is not None:
				if h is not None:
					self.context.SetPen(wx.Pen(wx.SystemSettings.GetColour(wx.SYS_COLOUR_SCROLLBAR)))
					self.context.SetBrush(wx.Brush(wx.SystemSettings.GetColour(wx.SYS_COLOUR_SCROLLBAR)))
					self.context.DrawRectangle(x + SCHEDULE_INSIDE_MARGIN, y + offsetY,
								   w - 2 * SCHEDULE_INSIDE_MARGIN, 10)
					if schedule.complete:
						self.context.SetPen(wx.Pen(wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT)))
						self.context.SetBrush(wx.Brush(wx.SystemSettings.GetColour(wx.SYS_COLOUR_HIGHLIGHT)))
						self.context.DrawRectangle(x + SCHEDULE_INSIDE_MARGIN, y + offsetY,
									   int((w - 2 * SCHEDULE_INSIDE_MARGIN) * schedule.complete), 10)

				offsetY += 10 + SCHEDULE_INSIDE_MARGIN

			if schedule.icons:
				for icon in schedule.icons:
					if h is not None:
						bitmap = wx.ArtProvider.GetBitmap( icon, wx.ART_FRAME_ICON, (16, 16) )
						self.context.DrawBitmap( bitmap, x + offsetX, y + offsetY, True )
					offsetX += 20
					if offsetX > w - SCHEDULE_INSIDE_MARGIN:
						offsetY += 20
						offsetX = SCHEDULE_INSIDE_MARGIN
						break

			font = schedule.font
			self.context.SetFont(font)

			self.context.SetTextForeground( schedule.foreground )
			offsetY += self._drawTextInRect( self.context, schedule.description, offsetX,
							 x, y + offsetY, w - 2 * SCHEDULE_INSIDE_MARGIN, None if h is None else h - offsetY - SCHEDULE_INSIDE_MARGIN )

		if h is not None:
			schedule.clientdata.bounds = (x, y, w, h)

		return offsetY

	def DrawScheduleVertical(self, schedule, day, workingHours, x, y, width, height):
		"""Draws a schedule vertically."""

		size, position, total = self.ScheduleSize(schedule, workingHours, day, 1)

		if self.use_gc:
			font = schedule.font
			self.context.SetFont(font, schedule.color)
		else:
			font = schedule.font
			self.context.SetTextForeground( schedule.foreground )
			self.context.SetFont(font)

		y = y + position * height / total + SCHEDULE_OUTSIDE_MARGIN
		x += SCHEDULE_OUTSIDE_MARGIN
		height = height * size / total - 2 * SCHEDULE_OUTSIDE_MARGIN
		width -= 2 * SCHEDULE_OUTSIDE_MARGIN

		self._DrawSchedule(schedule, x, y, width, height)
		return (x - SCHEDULE_OUTSIDE_MARGIN, y - SCHEDULE_OUTSIDE_MARGIN,
			width + 2 * SCHEDULE_OUTSIDE_MARGIN, height + 2 * SCHEDULE_OUTSIDE_MARGIN)

	def DrawScheduleHorizontal(self, schedule, day, daysCount, workingHours, x, y, width, height):
		"""Draws a schedule horizontally."""

		size, position, total = self.ScheduleSize(schedule, workingHours, day, daysCount)

		if self.use_gc:
			font = schedule.font
			self.context.SetFont(font, schedule.color)
		else:
			font = schedule.font
			self.context.SetTextForeground( schedule.color )
			self.context.SetFont(font)

		x = x + position * width / total + SCHEDULE_OUTSIDE_MARGIN
		width = width * size / total - 2 * SCHEDULE_OUTSIDE_MARGIN

		# Height is variable
		height = self._DrawSchedule( schedule, x, y, width, None )
		self._DrawSchedule(schedule, x, y, width, height)

		return (x - SCHEDULE_OUTSIDE_MARGIN, y - SCHEDULE_OUTSIDE_MARGIN,
			width + 2 * SCHEDULE_OUTSIDE_MARGIN, height + 2 * SCHEDULE_OUTSIDE_MARGIN)

	def ScheduleSize(schedule, workingHours, firstDay, dayCount):
		"""
		This convenience  static method computes  the position
		and size  size of the  schedule in the  direction that
		represent time,  according to a set  of working hours.
		The workingHours  parameter is  a list of  2-tuples of
		wx.DateTime  objects   defining  intervals  which  are
		indeed worked.  startPeriod and endPeriod  delimit the
		period.
		"""

		totalSpan = 0
		scheduleSpan = 0
		position = 0

		totalTime = 0
		for startHour, endHour in workingHours:
			totalTime += copyDateTime(endHour).Subtract(startHour).GetMinutes() / 60.0

		for dayNumber in xrange(dayCount):
			currentDay = copyDateTime(firstDay)
			currentDay.AddDS(wx.DateSpan(days=dayNumber))

			for startHour, endHour in workingHours:
				startHourCopy = wx.DateTimeFromDMY(currentDay.GetDay(),
								   currentDay.GetMonth(),
								   currentDay.GetYear(),
								   startHour.GetHour(),
								   startHour.GetMinute(),
								   0)
				endHourCopy = wx.DateTimeFromDMY(currentDay.GetDay(),
								 currentDay.GetMonth(),
								 currentDay.GetYear(),
								 endHour.GetHour(),
								 endHour.GetMinute(),
								 0)

				totalSpan += endHourCopy.Subtract(startHourCopy).GetMinutes()

				localStart = copyDateTime(schedule.start)

				if localStart.IsLaterThan(endHourCopy):
					position += endHourCopy.Subtract(startHourCopy).GetMinutes()
					continue

				if startHourCopy.IsLaterThan(localStart):
					localStart = startHourCopy

				localEnd = copyDateTime(schedule.end)

				if startHourCopy.IsLaterThan(localEnd):
					continue

				position += localStart.Subtract(startHourCopy).GetMinutes()

				if localEnd.IsLaterThan(endHourCopy):
					localEnd = endHourCopy

				scheduleSpan += localEnd.Subtract(localStart).GetMinutes()

		return dayCount * totalTime * scheduleSpan / totalSpan, dayCount * totalTime * position / totalSpan, totalTime * dayCount

	ScheduleSize = staticmethod(ScheduleSize)

	def _drawTextInRect( self, context, text, offsetX, x, y, w, h ):
		words = text.split()
		tw, th = context.GetTextExtent( u' '.join(words) )

		if h is not None and th > h + SCHEDULE_INSIDE_MARGIN:
			return SCHEDULE_INSIDE_MARGIN

		if tw <= w - offsetX:
			context.DrawText( u' '.join(words), x + offsetX, y )
			return th + SCHEDULE_INSIDE_MARGIN

		dpyWords = []
		remaining = w - offsetX
		totalW = 0
		spaceW, _ = context.GetTextExtent(u' ')

		for idx, word in enumerate(words):
			tw, _ = context.GetTextExtent(word)
			if remaining - tw - spaceW <= 0:
				break
			totalW += tw
			remaining -= tw + spaceW
			dpyWords.append(word)

		if dpyWords:
			words = words[idx:]

			currentX = 1.0 * offsetX
			if len(dpyWords) > 1:
				if words:
					spacing = (1.0 * (w - offsetX) - totalW) / (len(dpyWords) - 1)
				else:
					spacing = spaceW
			else:
				spacing = 0.0

			for word in dpyWords:
				tw, _ = context.GetTextExtent(word)
				context.DrawText(word, int(x + currentX), y)
				currentX += spacing + tw
		else:
			if offsetX == SCHEDULE_INSIDE_MARGIN:
				# Can't display anything...
				return SCHEDULE_INSIDE_MARGIN

		if words:
			ny = y + SCHEDULE_INSIDE_MARGIN + th
			if h is not None and ny > y + h:
				return SCHEDULE_INSIDE_MARGIN
			th += self._drawTextInRect( context, u' '.join(words), SCHEDULE_INSIDE_MARGIN, x, ny, w, None if h is None else h - (ny - y) )

		return th + SCHEDULE_INSIDE_MARGIN

	def _shrinkText( self, dc, text, width, height ):
		"""
		Truncate text at desired width
		"""
		MORE_SIGNAL		 = '...'
		SEPARATOR		 = " "

		textlist	 = list()	# List returned by this method
		words	 = list()	# Wordlist for itermediate elaboration

		# Split text in single words and split words when yours width is over 
		# available width
		text = text.replace( "\n", " " ).split()

		for word in text:
			if dc.GetTextExtent( word )[0] > width:
				# Cycle trought every char until word width is minor or equal
				# to available width
				partial = ""
				
				for char in word:
					if dc.GetTextExtent( partial + char )[0] > width:
						words.append( partial )
						partial = char
					else:
						partial += char
			else:
				words.append( word )

		# Create list of text lines for output
		textline = list()

		for word in words:
			if dc.GetTextExtent( SEPARATOR.join( textline + [word] ) )[0] > width:
				textlist.append( SEPARATOR.join( textline ) )
				textline = [word]

				# Break if there's no vertical space available
				if ( len( textlist ) * dc.GetTextExtent( SEPARATOR )[0] ) > height:
					# Must exists almost one line of description
					if len( textlist ) > 1:
						textlist = textlist[: - 1]

					break
			else:
				textline.append( word )

		# Add remained words to text list
		if len( textline ) > 0:
			textlist.append( SEPARATOR.join( textline ) )

		return textlist


class BackgroundDrawerDCMixin(object):
	"""
	Mixin to draw day background with a DC.
	"""

	def DrawDayBackground(self, x, y, w, h, highlight=None):
		if highlight is not None:
			self.context.SetBrush( wx.Brush( highlight ) )
		else:
			self.context.SetBrush( wx.TRANSPARENT_BRUSH )

		self.context.SetPen( FOREGROUND_PEN )

		self.context.DrawRectangle(x, y - 1, w, h + 1)


class HeaderDrawerDCMixin(object):
	"""
	A mixin to draw headers with a regular DC.
	"""

	def _DrawHeader(self, text, x, y, w, h, pointSize=12, weight=wx.FONTWEIGHT_BOLD,
			alignRight=False, highlight=None):
		font = self.context.GetFont()
		font.SetPointSize( pointSize )
		font.SetWeight( weight )
		self.context.SetFont( font )

		textW, textH = self.context.GetTextExtent( text )

		if highlight is not None:
			self.context.SetBrush( wx.Brush( highlight ) )
		else:
			self.context.SetBrush( wx.Brush( SCHEDULER_BACKGROUND_BRUSH() ) )

		self.context.DrawRectangle( x, y, w, textH * 1.5 )

		self.context.SetTextForeground( wx.BLACK )

		if alignRight:
			self.context.DrawText( text, x + w - textW * 1.5, y + textH * .25)
		else:
			self.context.DrawText( text, x + ( w - textW ) / 2, y + textH * .25 )

		return w, textH * 1.5

	def DrawSchedulesCompact(self, day, schedules, x, y, width, height, highlightColor):
		if day is None:
			self.context.SetBrush(wx.LIGHT_GREY_BRUSH)
		else:
			self.context.SetBrush(wx.Brush(DAY_BACKGROUND_BRUSH()))

		self.context.DrawRectangle(x, y, width, height)

		results = []

		if day is not None:
			if day.IsSameDate(wx.DateTime.Now()):
				color = highlightColor
			else:
				color = None
			headerW, headerH = self.DrawSimpleDayHeader(day, x, y, width, height,
								    highlight=color)
			y += headerH
			height -= headerH

			x += SCHEDULE_OUTSIDE_MARGIN
			width -= 2 * SCHEDULE_OUTSIDE_MARGIN

			y += SCHEDULE_OUTSIDE_MARGIN
			height -= 2 * SCHEDULE_OUTSIDE_MARGIN

			self.context.SetPen(FOREGROUND_PEN)

			totalHeight = 0

			for schedule in schedules:
				if schedule.start.Format('%H%M') != '0000':
					description = '%s %s' % (wxTimeFormat.FormatTime(schedule.start, includeMinutes=True), schedule.description)
				else:
					description = schedule.description
				description = self._shrinkText(self.context, description, width - 2 * SCHEDULE_INSIDE_MARGIN, headerH)[0]

				textW, textH = self.context.GetTextExtent(description)
				if totalHeight + textH > height:
					break

				self.context.SetBrush(wx.Brush(schedule.color))
				self.context.DrawRectangle(x, y, width, textH * 1.2)
				results.append((schedule, wx.Point(x, y), wx.Point(x + width, y + textH * 1.2)))

				self.context.SetTextForeground(schedule.foreground)
				self.context.DrawText(description, x + SCHEDULE_INSIDE_MARGIN, y + textH * 0.1)

				y += textH * 1.2
				totalHeight += textH * 1.2

		return results


class BackgroundDrawerGCMixin(object):
	"""
	Mixin to draw day background with a GC.
	"""

	def DrawDayBackground(self, x, y, w, h, highlight=None):
		if highlight is not None:
			self.context.SetBrush( self.context.CreateLinearGradientBrush( x, y, x + w, y + h,
										       wx.Colour(128, 128, 128, 128),
										       wx.Colour(highlight.Red(), highlight.Green(), highlight.Blue(), 128) ) )
		else:
			self.context.SetBrush( self.context.CreateBrush( wx.TRANSPARENT_BRUSH ) )

		self.context.SetPen( self.context.CreatePen( FOREGROUND_PEN ) )

		self.context.DrawRectangle(x, y - 1, w, h + 1)


class HeaderDrawerGCMixin(object):
	"""
	A mixin to draw headers with a GraphicsContext.
	"""

	def _DrawHeader(self, text, x, y, w, h, pointSize=12, weight=wx.FONTWEIGHT_BOLD,
			alignRight=False, highlight=None):
		font = wx.NORMAL_FONT
		fsize = font.GetPointSize()
		fweight = font.GetWeight()

		try:
			font.SetPointSize( pointSize )
			font.SetWeight( weight )
			self.context.SetFont(font, wx.BLACK)

			textW, textH = self.context.GetTextExtent( text )

			x1 = x
			y1 = y
			x2 = x + w
			y2 = y + textH * 1.5

			if highlight is not None:
				self.context.SetBrush(self.context.CreateLinearGradientBrush(x1, y1, x2, y2, wx.Colour(128, 128, 128),
											     highlight))
			else:
				self.context.SetBrush(self.context.CreateLinearGradientBrush(x1, y1, x2, y2, wx.Colour(128, 128, 128),
											     SCHEDULER_BACKGROUND_BRUSH()))
			self.context.DrawRectangle(x1, y1, x2 - x1, y2 - y1)

			if alignRight:
				self.context.DrawText(text, x + w - 1.5 * textW, y + textH * .25)
			else:
				self.context.DrawText(text, x + (w - textW) / 2, y + textH * .25)

			return w, textH * 1.5
		finally:
			font.SetPointSize(fsize)
			font.SetWeight(fweight)

	def DrawSchedulesCompact(self, day, schedules, x, y, width, height, highlightColor):
		if day is None:
			brush = self.context.CreateLinearGradientBrush(x, y, x + width, y + height, wx.BLACK, SCHEDULER_BACKGROUND_BRUSH())
		else:
			brush = self.context.CreateLinearGradientBrush(x, y, x + width, y + height, wx.LIGHT_GREY, DAY_BACKGROUND_BRUSH())

		self.context.SetBrush(brush)
		self.context.DrawRectangle(x, y, width, height)

		font = wx.NORMAL_FONT
		fsize = font.GetPointSize()
		fweight = font.GetWeight()

		try:
			font.SetPointSize(10)
			font.SetWeight(wx.FONTWEIGHT_NORMAL)

			results = []

			if day is not None:
				if day.IsSameDate(wx.DateTime.Now()):
					color = highlightColor
				else:
					color = None
				headerW, headerH = self.DrawSimpleDayHeader(day, x, y, width, height,
									    highlight=color)
				y += headerH
				height -= headerH

				x += SCHEDULE_OUTSIDE_MARGIN
				width -= 2 * SCHEDULE_OUTSIDE_MARGIN

				y += SCHEDULE_OUTSIDE_MARGIN
				height -= 2 * SCHEDULE_OUTSIDE_MARGIN

				self.context.SetPen(FOREGROUND_PEN)

				totalHeight = 0

				for schedule in schedules:
					if schedule.start.Format('%H%M') != '0000':
						description = '%s %s' % (wxTimeFormat.FormatTime(schedule.start, includeMinutes=True), schedule.description)
					else:
						description = schedule.description
					description = self._shrinkText(self.context, description, width - 2 * SCHEDULE_INSIDE_MARGIN, headerH)[0]

					textW, textH = self.context.GetTextExtent(description)
					if totalHeight + textH > height:
						break

					brush = self.context.CreateLinearGradientBrush(x, y, x + width, y + height, schedule.color, DAY_BACKGROUND_BRUSH())
					self.context.SetBrush(brush)
					self.context.DrawRoundedRectangle(x, y, width, textH * 1.2, 1.0 * textH / 2)
					results.append((schedule, wx.Point(x, y), wx.Point(x + width, y + textH * 1.2)))

					self.context.SetFont(schedule.font, schedule.foreground)
					self.context.DrawText(description, x + SCHEDULE_INSIDE_MARGIN, y + textH * 0.1)

					y += textH * 1.2
					totalHeight += textH * 1.2

			return results
		finally:
			font.SetPointSize(fsize)
			font.SetWeight(fweight)


class HeaderDrawerMixin(object):
	"""
	A mixin that draws header using the _DrawHeader method.
	"""

	def DrawDayHeader(self, day, x, y, width, height, highlight=None):
		return self._DrawHeader('%s %s %s' % ( day.GetWeekDayName( day.GetWeekDay() )[:3],
						       day.GetDay(), day.GetMonthName( day.GetMonth() ) ),
					x, y, width, height, highlight=highlight)

	def DrawMonthHeader(self, day, x, y, w, h):
		return self._DrawHeader('%s %s' % ( day.GetMonthName( day.GetMonth() ), day.GetYear() ),
					x, y, w, h)

	def DrawSimpleDayHeader(self, day, x, y, w, h, highlight=None):
		return self._DrawHeader(day.Format('%a %d'), x, y, w, h,
					weight=wx.FONTWEIGHT_NORMAL, alignRight=True,
					highlight=highlight)


class wxBaseDrawer(BackgroundDrawerDCMixin, HeaderDrawerDCMixin, HeaderDrawerMixin, wxDrawer):
	"""
	Concrete subclass of wxDrawer; regular style.
	"""

	def DrawHours(self, x, y, w, h, direction, includeText=True):
		if direction == wxSCHEDULER_VERTICAL:
			self.context.SetBrush(wx.Brush(SCHEDULER_BACKGROUND_BRUSH()))
			self.context.DrawRectangle(x, y, LEFT_COLUMN_SIZE, h)

		font = self.context.GetFont()
		fWeight = font.GetWeight()
		fSize = font.GetPointSize()
		try:
			font.SetWeight( wx.FONTWEIGHT_NORMAL )
			self.context.SetFont( font )
			self.context.SetTextForeground( wx.BLACK )

			if direction == wxSCHEDULER_VERTICAL:
				hourH = 1.0 * h / len(self.displayedHours)
				self.AdjustFontForHeight( font, hourH )
				hourW, _ = self.context.GetTextExtent( ' ' + wxTimeFormat.FormatTime( wx.DateTimeFromHMS(23, 59, 59) ) )
			else:
				hourW = 1.0 * w / len(self.displayedHours)
				self.AdjustFontForWidth( font, int(hourW * 2 * 0.9) )
				_, hourH = self.context.GetTextExtent( ' ' + wxTimeFormat.FormatTime( wx.DateTimeFromHMS(23, 59, 59) ) )

			if not includeText:
				hourH = 0

			for i, hour in enumerate( self.displayedHours ):
				if hour.GetMinute() == 0:
					if direction == wxSCHEDULER_VERTICAL:
						self.context.DrawLine(x + LEFT_COLUMN_SIZE - hourW / 2, y + i * hourH, x + w, y + i * hourH)
						if includeText:
							self.context.DrawText(wxTimeFormat.FormatTime(hour), x + LEFT_COLUMN_SIZE - hourW - 5, y + i * hourH)
					else:
						self.context.DrawLine(x + i * hourW, y + hourH * 1.25, x + i * hourW, y + h)
						if includeText:
							self.context.DrawText(wxTimeFormat.FormatTime(hour), x + i * hourW + 5, y + hourH * .25)
				else:
					if direction == wxSCHEDULER_VERTICAL:
						self.context.DrawLine(x + LEFT_COLUMN_SIZE, y + i * hourH, x + w, y + i * hourH)
					else:
						self.context.DrawLine(x + i * hourW, y + hourH * 1.4, x + i * hourW, y + h)

			if direction == wxSCHEDULER_VERTICAL:
				self.context.DrawLine(x + LEFT_COLUMN_SIZE - 1, y, x + LEFT_COLUMN_SIZE - 1, y + h)
				return LEFT_COLUMN_SIZE, max(h, DAY_SIZE_MIN.height)
			else:
				self.context.DrawLine(x, y + hourH * 1.5 - 1, x + w, y + hourH * 1.5 - 1)
				return max(w, DAY_SIZE_MIN.width), hourH * 1.5
		finally:
			font.SetWeight( fWeight )
			font.SetPointSize( fSize )

	def DrawNowHorizontal(self, x, y, w):
		self.context.SetBrush( wx.Brush( wx.Colour( 0, 128, 0 ) ) )
		self.context.SetPen( wx.Pen( wx.Colour( 0, 128, 0 ) ) )
		self.context.DrawArc( x, y + 5, x, y - 5, x, y )
		self.context.DrawRectangle( x, y - 1, w, 3 )

	def DrawNowVertical(self, x, y, h):
		self.context.SetBrush( wx.Brush( wx.Colour( 0, 128, 0 ) ) )
		self.context.SetPen( wx.Pen( wx.Colour( 0, 128, 0 ) ) )
		self.context.DrawArc( x - 5, y, x + 5, y, x, y )
		self.context.DrawRectangle( x - 1, y, 3, h )


class wxFancyDrawer(BackgroundDrawerGCMixin, HeaderDrawerGCMixin, HeaderDrawerMixin, wxDrawer):
	"""
	Concrete subclass of wxDrawer; fancy eye-candy using wx.GraphicsContext.
	"""

	use_gc = True

	def DrawHours(self, x, y, w, h, direction, includeText=True):
		if direction == wxSCHEDULER_VERTICAL:
			brush = self.context.CreateLinearGradientBrush(x, y, x + w, y + h, SCHEDULER_BACKGROUND_BRUSH(), DAY_BACKGROUND_BRUSH())
			self.context.SetBrush(brush)
			self.context.DrawRectangle(x, y, LEFT_COLUMN_SIZE, h)

		font = wx.NORMAL_FONT
		fsize = font.GetPointSize()
		fweight = font.GetWeight()

		try:
			font.SetWeight(wx.FONTWEIGHT_NORMAL)
			self.context.SetFont(font, wx.BLACK)

			self.context.SetPen(FOREGROUND_PEN)

			if direction == wxSCHEDULER_VERTICAL:
				hourH = 1.0 * h / len(self.displayedHours)
				self.AdjustFontForHeight( font, hourH )
				hourW, _ = self.context.GetTextExtent( ' ' + wxTimeFormat.FormatTime( wx.DateTimeFromHMS(23, 59, 59) ) )
			else:
				hourW = 1.0 * w / len(self.displayedHours)
				self.AdjustFontForWidth( font, int(hourW * 2 * 0.9) )
				_, hourH = self.context.GetTextExtent( ' ' + wxTimeFormat.FormatTime( wx.DateTimeFromHMS(23, 59, 59) ) )

			if not includeText:
				hourH = 0

			for i, hour in enumerate( self.displayedHours ):
				if hour.GetMinute() == 0:
					if direction == wxSCHEDULER_VERTICAL:
						self.context.DrawLines([(x + LEFT_COLUMN_SIZE - hourW / 2, y + i * hourH),
									(x + w, y + i * hourH)])
						if includeText:
							self.context.DrawText(' ' + wxTimeFormat.FormatTime(hour), x + LEFT_COLUMN_SIZE - hourW - 10, y + i * hourH)
					else:
						self.context.DrawLines([(x + i * hourW, y + hourH * 1.25),
									(x + i * hourW, y + h + 10)])
						if includeText:
							self.context.DrawText(wxTimeFormat.FormatTime(hour), x + i * hourW + 5, y + hourH * .25)
				else:
					if direction == wxSCHEDULER_VERTICAL:
						self.context.DrawLines([(x + LEFT_COLUMN_SIZE, y + i * hourH), (x + w, y + i * hourH)])
					else:
						self.context.DrawLines([(x + i * hourW, y + hourH * 1.4), (x + i * hourW, y + h)])

			if direction == wxSCHEDULER_VERTICAL:
				self.context.DrawLines([(x + LEFT_COLUMN_SIZE - 1, y),
							(x + LEFT_COLUMN_SIZE - 1, y + h)])
				return LEFT_COLUMN_SIZE, max(h, DAY_SIZE_MIN.height)
			else:
				self.context.DrawLines([(x, y + hourH * 1.5 - 1), (x + w, y + hourH * 1.5 - 1)])
				return max(w, DAY_SIZE_MIN.width), hourH * 1.5
		finally:
			font.SetPointSize( fsize )
			font.SetWeight( fweight )

	def DrawNowHorizontal(self, x, y, w):
		brush = self.context.CreateLinearGradientBrush( x + 4, y - 1, x + w, y + 1, wx.Colour( 0, 128, 0, 128 ), wx.Colour( 0, 255, 0, 128 ) )
		self.context.SetBrush( brush )
		self.context.DrawRectangle( x + 4, y - 2, w - 4, 3 )

		brush = self.context.CreateRadialGradientBrush( x, y - 5, x, y, 5, wx.Colour( 0, 128, 0, 128 ), wx.Colour( 0, 255, 0, 128 ) )
		self.context.SetBrush( brush )

		path = self.context.CreatePath()
		path.AddArc( x, y, 5, -math.pi / 2, math.pi / 2, True )
		self.context.FillPath( path )

	def DrawNowVertical(self, x, y, h):
		brush = self.context.CreateLinearGradientBrush( x - 1, y + 4, x + 1, y + h, wx.Colour( 0, 128, 0, 128 ), wx.Colour( 0, 255, 0, 128 ) )
		self.context.SetBrush( brush )
		self.context.DrawRectangle( x - 2, y + 4, 3, h - 4 )

		brush = self.context.CreateRadialGradientBrush( x - 5, y, x, y, 5, wx.Colour( 0, 128, 0, 128 ), wx.Colour( 0, 255, 0, 128 ) )
		self.context.SetBrush(brush)

		path = self.context.CreatePath()
		path.AddArc( x, y, 5, 0.0, math.pi, True )
		self.context.FillPath( path )



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

import generator, os
from taskcoachlib import meta


css = '''/*
CSS file generated by %(name)s %(version)s for:
%%s. 
You can edit this file, it will not be overwritten by %(name)s.
*/
'''%meta.data.metaDict + generator.css + '''
/* Other possibilities to tune the layout include:

   Change the styles for a header of a specific column, in this case the subject 
   column. Note how this style overrides the default style in the HTML file: 
   
   th.subject {
      text-align: center;
   }

   If the exported HTML file contains tasks it is possible to change the color 
   of completed (or overdue, duesoon, late, inactive, active) tasks like this:
   
   .completed {
       color: red;
   }
   
*/
'''

class HTMLWriter(object):
    def __init__(self, fd, filename=None):
        self.__fd = fd
        self.__filename = filename
        self.__cssFilename = os.path.splitext(filename)[0] + '.css' if filename else ''

    def write(self, viewer, settings, selectionOnly=False, separateCSS=False, columns=None):
        cssFilename = os.path.basename(self.__cssFilename) if separateCSS else ''
        htmlText, count = generator.viewer2html(viewer, settings, cssFilename, 
                                                selectionOnly, columns)
        self.__fd.write(htmlText)
        if separateCSS:
            self._writeCSS()
        return count
    
    def _writeCSS(self, open=open): # pylint: disable=W0622
        if not self.__cssFilename or os.path.exists(self.__cssFilename):
            return
        try:
            fd = open(self.__cssFilename, 'wb')
            fd.write(css%self.__filename)
            fd.close()
        except IOError:
            pass



'''
Task Coach - Your friendly task manager
Copyright (C) 2004-2016 Task Coach developers <developers@taskcoach.org>

Task Coach is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Task Coach is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program.  If not, see <http://www.gnu.org/licenses/>.
'''

from notebook import Notebook, BookPage
from frame import AuiManagedFrameWithDynamicCenterPane
from dialog import Dialog, NotebookDialog, HTMLDialog, AttachmentSelector
from itemctrl import Column
from listctrl import VirtualListCtrl
from checklistbox import CheckListBox
from treectrl import CheckTreeCtrl, TreeListCtrl
from squaremap import SquareMap
from timeline import Timeline
from datectrl import DateTimeCtrl, TimeEntry
from textctrl import SingleLineTextCtrl, MultiLineTextCtrl, StaticTextWithToolTip
from panel import PanelWithBoxSizer, BoxWithFlexGridSizer, BoxWithBoxSizer
from searchctrl import SearchCtrl
from spinctrl import SpinCtrl
from tooltip import ToolTipMixin, SimpleToolTip
from dirchooser import DirectoryChooser
from fontpicker import FontPickerCtrl
from syncmlwarning import SyncMLWarningDialog
from calendarwidget import Calendar
from calendarconfig import CalendarConfigDialog
from password import GetPassword
from hcalendar import HierarchicalCalendar
from hcalendarconfig import HierarchicalCalendarConfigDialog
import masked
from wx.lib import sized_controls
